from typing import List, Dict, Any, Optional, Callable

from versionhq import Product, ProductProvider, Customer
from versionhq.clients.customer.model import  Status


def test_create_customer():
    provider = ProductProvider(name="demo", region="US", data_pipelines=["hubspot", "demo crm"], destination_services=["test"])
    product = Product(name="demo", description="demo", provider=provider, landing_url="www.com", cohort_timeframe=30)

    def custom_deploy(item):
        return f"custom deploy with {item}"

    class CustomCustomer(Customer):
        name: str = "custom"
        staus: Status = Status.READY_TO_DEPLOY
        products: List[Product] = [product,]
        analysis: str = "analysis"
        function: Optional[Callable] = custom_deploy
        config: Optional[Dict[str, Any]] = dict(item="custom config")

    customer = CustomCustomer()

    assert customer.id is not None
    assert customer.deploy() == "custom deploy with custom config"
    assert customer.status == Status.ACTIVE_ON_WORKFLOW
    assert [isinstance(item, Product) for item in customer.products]



from unittest.mock import patch
from typing import Callable

from pydantic import BaseModel

from versionhq.agent.model import Agent
from versionhq.llm.model import LLM
from versionhq.task.model import Task, ResponseField, TaskOutput
from versionhq.tool.model import Tool, ToolSet
from versionhq.tool.decorator import tool


class CustomOutput(BaseModel):
    test1: str
    test2: list[str]

def mock_tool() -> str:
    return "mock"


def test_store_task_log():
    task = Task(
        description="return the output following the given prompt.",
        response_schema=[ResponseField(title="task_1", data_type=str, required=True),],
    )
    task.execute()

    from versionhq.storage.task_output_storage import TaskOutputStorageHandler
    assert TaskOutputStorageHandler().load() is not None


def test_task_with_agent_tools():
    simple_tool = Tool(name="simple tool", func=lambda x: "simple func")
    agent = Agent(role="demo", goal="execute tools", tools=[simple_tool,], maxit=1, max_tokens=3000)
    task = Task(description="execute tool", can_use_agent_tools=True, tool_res_as_final=True)
    res = task.execute(agent=agent)
    assert res.tool_output == "simple func"

    def empty_func():
        return "empty func"

    func_tool = Tool(name="func tool", func=empty_func)
    agent.tools = [func_tool]
    res = task.execute(agent=agent)
    assert res.tool_output == "empty func"


    def demo_func(message: str) -> str:
        return message + "_demo"

    class CustomTool(Tool):
        name: str = "custom tool"

    custom_tool = CustomTool(func=demo_func)
    agent.tools = [custom_tool]
    res = task.execute(agent=agent)
    assert "_demo" in res.tool_output


def test_task_with_tools():

    def random_func(message: str) -> str:
        return message + "_demo"

    tool = Tool(name="tool", func=random_func)
    tool_set = ToolSet(tool=tool, kwargs=dict(message="empty func"))
    task = Task(description="execute the given tools", tools=[tool_set,], tool_res_as_final=True)
    # res = task.execute()
    # assert res.tool_output == "empty func_demo"

    class CustomTool(Tool):
        name: str = "custom tool"
        func: Callable = None

    custom_tool = CustomTool(func=random_func)
    task.tools = [custom_tool]
    res = task.execute()
    assert "_demo" in res.tool_output

    task.tools = [custom_tool]
    res = task.execute()
    assert res.tool_output is not None


def test_task_without_response_format():
    task = Task(description="return a simple output with any random values.")
    res = task.execute()

    assert res and isinstance(res, TaskOutput)
    assert res.json_dict and isinstance(res.json_dict, dict)
    assert res.pydantic is None


def test_build_agent_without_developer_prompt():
    agent = Agent(
        role="analyst",
        goal="analyze the company's website and retrieve the product overview",
        backstory="You are competitive analysts who have abundand knowledge in marketing, product management.",
        use_developer_prompt=False
    )
    task = Task(description="return a simple output with any random values.")
    res = task.execute(agent=agent)

    assert res and isinstance(res, TaskOutput)
    assert res.json_dict and isinstance(res.json_dict, dict)
    assert res.pydantic is None


def test_task_with_agent_callback():
    import litellm

    def mock_func() -> str:
        return "Demo func"

    agent = Agent(role="demo", goal="amazing project goal", maxit=1, max_tokens=3000, callbacks=[mock_func,])
    task = Task(description="Amazing task")
    res = task.execute(agent=agent)

    assert res.raw and res.task_id == task.id
    assert litellm.callbacks == [mock_func]


def test_rpm():
    agent = Agent(role="demo", goal="use the given tools", max_tokens=3000, max_rpm=3)
    assert agent._rpm_controller and agent._rpm_controller.max_rpm == 3

    a = "hello"
    tool = Tool(func=lambda x: a + x)
    tool_set = ToolSet(tool=tool, kwargs={ "x": "_demo" })
    task = Task(description="Summarize overview of the given tool in sentence, then execute the tool.", tools=[tool_set,])
    res = task.execute(agent=agent)

    assert "hello_demo" in res.raw if res.raw else res
    assert agent._rpm_controller._current_rpm < agent._rpm_controller.max_rpm


def test_maxit():
    @tool
    def demo() -> str:
        """Get the final answer but don't give it yet, just re-use this
        tool non-stop."""
        return "demo"

    agent = Agent(role="demo", goal="amazing demo", maxit=2)
    task = Task(description="Summarize overview of the given tool in sentences.", tools=[demo,])

    with patch.object(LLM, "call", wraps=agent.llm.call) as mock:
        task.execute(agent=agent)
        assert mock.call_count <= 2


def test_taskoutput_final_a():
    task = Task(
        description="Research any topic.",
        response_schema=CustomOutput,
        tools=[mock_tool],
        tool_res_as_final=True
    )
    res = task.execute()

    assert res.final == res.tool_output
    assert res._to_context_prompt() is not None


def test_taskoutput_final_b():
    task = Task(
        description="Research any topic.",
        tools=[mock_tool],
        tool_res_as_final=False
    )
    res = task.execute()

    assert res.final == res.json_dict
    assert res._to_context_prompt() is not None


def test_taskoutput_final_c():
    task = Task(description="Research any topic.")
    res = task.execute()

    assert res.final == res.json_dict
    assert res._to_context_prompt() is not None


def test_taskoutput_final_c():
    task = Task(description="Research any topic.",  response_schema=CustomOutput)
    res = task.execute()

    assert res.final == res.pydantic
    assert res._to_context_prompt() is not None


def test_multimodal():
    import versionhq as vhq

    agent = vhq.Agent(llm="azure/whisper", role="Transcripter")
    task = vhq.Task(description="trasncript the audio", audio='tests/_sample/sample.mp3', is_multimodal=True)
    res = task.execute(agent=agent)
    assert res.raw is not None



from enum import Enum


class GPTSizeEnum(str, Enum):
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"


class GPTCUABrowserEnum(str, Enum):
    BROWSER = "browser"
    MAC = "mac"
    WINDOWS = "windows"
    UNBUNTU = "ubuntu"


class GPTCUATypeEnum(str, Enum):
    COMPUTER_CALL_OUTPUT = "computer_call_output"
    COMPUTER_USE_PREVIEW = "computer_use_preview"


class GPTFilterTypeEnum(str, Enum):
    eq = "eq"
    ne = "ne"
    gt = "gt"
    gte = "gte"
    lt = "lt"
    lte = "lte"



import matplotlib
matplotlib.use('agg')

import uuid
import datetime
from enum import IntEnum, Enum
from abc import ABC
from concurrent.futures import Future
from typing import List, Any, Optional, Callable, Dict, Type, Tuple
from typing_extensions import Self

import networkx as nx
import matplotlib.pyplot as plt
from pydantic import BaseModel, InstanceOf, Field, UUID4, field_validator, model_validator
from pydantic_core import PydanticCustomError

from versionhq.agent.model import Agent
from versionhq.task.model import Task, TaskOutput, Evaluation, ResponseField
from versionhq._utils import Logger, UsageMetrics, ErrorType


class ReformTriggerEvent(IntEnum):
    USER_INPUT = 1 # ask human
    TEST_TIME_COMPUTATION = 2 # mismatch between actual responses and expected outcome
    ERROR_DETECTION = 3 # response error


class ConditionType(IntEnum):
    AND = 1
    OR = 2


class Condition(BaseModel):
    """
    A Pydantic class to store edge conditions and their args and types.
    """
    # edge_id: UUID4 = uuid.uuid4()
    methods: Dict[str, Callable | "Condition"] = dict()
    args: Dict[str, Dict[str, Any]] = dict()
    type: ConditionType = None

    @model_validator(mode="after")
    def validate_type(self) -> Self:
        if len(self.methods.keys()) > 1 and self.type is None:
            raise PydanticCustomError("missing_type", "Missing type", {})
        return self

    def _execute_method(self, key: str, method: Callable | "Condition") -> bool:
        match method:
            case Condition():
                return method.condition_met()
            case _:
                args = self.args[key] if key in self.args else None
                res = method(**args) if args else method()
                return res

    def condition_met(self) -> bool:
        if not self.methods:
            return True

        if len(self.methods) == 1:
            for k, v in self.methods.items():
               return self._execute_method(key=k, method=v)
        else:
            cond_list = []
            for k, v in self.methods.items():
                res = self._execute_method(key=k, method=v)
                if self.type == ConditionType.OR and res == True:
                    return True
                elif self.type == ConditionType.AND and res == False:
                    return False
            return bool(len([item for item in cond_list if item == True]) == len(cond_list))


class TaskStatus(IntEnum):
    """
    Enum to track the task execution status
    """
    NOT_STARTED = 1
    IN_PROGRESS = 2
    WAITING = 3     # waiting for its dependant tasks to complete. resumption set as AUTO.
    COMPLETED = 4
    DELAYED = 5     # task in progress - but taking longer than expected duration
    ON_HOLD = 6     # intentionally paused due to external factors and decisions. resumption set as DECISION.
    ERROR = 7       # tried task execute but returned error. resupmtion follows edge weights and agent settings


class DependencyType(str, Enum):
    """
    Concise enumeration of the edge type.
    """
    FINISH_TO_START = "FS"  # Task B starts after Task A finishes
    START_TO_START = "SS"  # Task B starts when Task A starts
    FINISH_TO_FINISH = "FF"  # Task B finishes when Task A finishes
    START_TO_FINISH = "SF"  # Task B finishes when Task A starts


class Node(BaseModel):
    """
    A class to store a node object.
    """

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    task: InstanceOf[Task] = Field(default=None)
    in_degree_nodes: List["Node"] = Field(default_factory=list, description="list of Node objects")
    out_degree_nodes: List["Node"] = Field(default_factory=list, description="list of Node objects")
    assigned_to: InstanceOf[Agent] = Field(default=None)
    status: TaskStatus = Field(default=TaskStatus.NOT_STARTED)

    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise PydanticCustomError("may_not_set_field", "This field is not to be set by client.", {})

    def is_independent(self) -> bool:
        return not self.in_degree_nodes and not self.out_degree_nodes

    def handle_task_execution(self, agent: Agent = None, context: str = None, response_format: Type[BaseModel] = None) -> TaskOutput | None:
        """Executes the task and updates its status"""

        if not self.task:
            Logger().log(level="error", message="Missing a task to execute. We'll return None.", color="red")
            self.status = TaskStatus.ERROR
            return None

        if self.status == TaskStatus.COMPLETED:
            return self.task.output

        else:
            self.status = TaskStatus.IN_PROGRESS
            agent = agent if agent else self.assigned_to
            self.task.response_schema = self.task.response_schema if self.task.response_schema else response_format if type(response_format) == BaseModel or isinstance(response_format, list) else None
            res = self.task.execute(agent=agent, context=context)

            if isinstance(res, Future): # activate async
                res = res.result()

            self.status = TaskStatus.COMPLETED if res else TaskStatus.ERROR
            return res

    @property
    def in_degrees(self) -> int:
        return len(self.in_degree_nodes) if self.in_degree_nodes else 0

    @property
    def out_degrees(self) -> int:
        return len(self.out_degree_nodes) if self.out_degree_nodes else 0

    @property
    def degrees(self) -> int:
        return self.in_degrees + self.out_degrees

    @property
    def identifier(self) -> str:
        """Unique identifier of the node"""
        return f"{str(self.id)}"

    @property
    def label(self) -> str:
        """Human friendly label for visualization"""
        return self.task.name if self.task.name else self.task.description[0: 8]

    def __str__(self):
        if self.task:
            return f"{self.identifier}: {self.task.name if self.task.name else self.task.description[0: 12]}"
        else:
            return self.identifier


class Edge(BaseModel):
    """
    A Pydantic class to store an edge object that connects source and target nodes.
    """

    source: Node = Field(default=None)
    target: Node = Field(default=None)

    description: Optional[str] = Field(default=None)
    weight: Optional[float | int] = Field(default=1, description="est. duration of the task execution or respective weight of the target node at any scale i.e., 1 low - 10 high priority")
    dependency_type: DependencyType = Field(default=DependencyType.FINISH_TO_START)
    required: bool = Field(default=True, description="whether to consider the source's status")
    condition: Optional[Condition] = Field(default=None)

    lag: Optional[float | int] = Field(default=None, description="lag time (sec) from the dependency met to the task execution")
    data_transfer: bool = Field(default=True, description="whether the data transfer is required. by default transfer plane text output from in-degree nodes as context")


    def _response_schema(self) -> Type[BaseModel]:
        class EdgeResponseSchema(BaseModel):
            wight: float
            dependecy_type: str
            required: bool
            need_condition: bool
            lag_in_sec: float

        return EdgeResponseSchema


    def dependency_met(self) -> bool:
        """
        Defines if the dependency is ready to execute:

        required - condition - Dependency Met? - Dependent Task Can Start?
        True	Not Given	Predecessor task finished	Yes
        True	Given	Predecessor task finished and condition True	Yes
        False	Not Given	Always (regardless of predecessor status)	Yes
        False	Given	Condition True (predecessor status irrelevant)	Yes
        """

        if self.required == False:
            return self.condition.condition_met() if self.condition else True
            # return self.condition(**self.condition_kwargs) if self.condition else True

        match self.dependency_type:
            case DependencyType.FINISH_TO_START:
                """target starts after source finishes"""
                if not self.source or self.source.status == TaskStatus.COMPLETED:
                    return self.condition.condition_met() if self.condition else True
                else:
                    return False

            case DependencyType.START_TO_START:
                """target starts when source starts"""
                if not self.source or self.source.status != TaskStatus.NOT_STARTED:
                    return self.condition.condition_met() if self.condition else True
                else:
                    return False

            case DependencyType.FINISH_TO_FINISH:
                """target finish when source start"""
                if not self.source or self.source.status != TaskStatus.COMPLETED:
                    return self.condition.condition_met() if self.condition else True
                else:
                    return False

            case DependencyType.START_TO_FINISH:
                """target finishes when source start"""
                if not self.source or self.source.status == TaskStatus.IN_PROGRESS:
                    return self.condition.condition_met() if self.condition else True
                else:
                    return False


    def activate(self, response_format: Type[BaseModel] = None) -> TaskOutput | None:
        """
        Activates the edge to initiate task execution of the target node.
        """
        if not self.source or not self.target:
            Logger(verbose=True).log(level="warning", message="Cannot find source or target nodes. We'll return None.", color="yellow")
            return None

        if not self.dependency_met():
            Logger(verbose=True).log(level="warning", message="Dependencies not met. We'll return None.", color="yellow")
            return None

        if self.lag:
            import time
            time.sleep(self.lag)

        context = self.source.task.output._to_context_prompt() if self.data_transfer else None
        res = self.target.handle_task_execution(context=context, response_format=response_format)
        return res


    @property
    def label(self):
        """Human friendly label for visualization."""
        return f"e{self.source.label}-{self.target.label}"


class Graph(ABC, BaseModel):
    """
    An abstract class to store G using NetworkX library.
    """
    directed: bool = Field(default=False, description="Whether the graph is directed")
    graph: Type[nx.Graph] = Field(default=None)
    nodes: Dict[str, Node] = Field(default_factory=dict, description="{node_identifier: Node}")
    edges: Dict[str, Edge] = Field(default_factory=dict)

    def __init__(self, directed: bool = False, **kwargs):
        super().__init__(directed=directed, **kwargs)
        self.graph = nx.DiGraph(directed=True) if self.directed else nx.Graph()

    def _return_node_object(self, node_identifier) -> Node | None:
        match = [v for k, v in self.nodes.items() if k == node_identifier]

        if match:
            node = match[0] if isinstance(match[0], Node) else match[0]["node"] if "node" in match[0] else None
            return node
        else:
            return None

    def _format_Graph(self) -> None:
        """Formats dxGraph using edges and nodes."""
        edges, nodes = [k for k in self.edges.keys()], [k for k in self.nodes.keys()]
        if self.graph:
            self.graph.update(edges=edges, nodes=nodes)
        else:
            self.graph = nx.Graph(directed=self.directed, edges=edges, nodes=nodes)

    def add_node(self, node: Node) -> None:
        if node.identifier in self.nodes.keys():
            return
        self.graph.add_node(node.identifier, node=node)
        self.nodes[node.identifier] = node

    def add_edge(self, source: str, target: str, edge: Edge) -> None:
        self.graph.add_edge(source, target, edge=edge)

        source_node, target_node = self._return_node_object(source), self._return_node_object(target)

        edge.source = source_node
        source_node.out_degree_nodes.append(target_node)

        edge.target = target_node
        target_node.in_degree_nodes.append(source_node)

        self.edges[(source, target)] = edge

    def add_weighted_edges_from(self, edges):
        self.graph.add_weighted_edges_from(edges)

    def get_neighbors(self, node: Node) -> List[Node]:
        return list(self.graph.neighbors(node))

    def get_in_degree(self, node: Node) -> int:
        return self.graph.in_degree(node)

    def get_out_degree(self, node: Node) -> int:
        return self.graph.out_degree(node)

    def find_start_nodes(self) -> List[Node]:
        return [v for v in self.nodes.values() if v.in_degrees == 0 and v.out_degrees > 0]

    def find_end_nodes(self) ->  List[Node]:
        return [v for v in self.nodes.values() if v.out_degrees == 0 and v.in_degrees > 0]

    def find_critical_end_node(self) -> Node | None:
        """Finds a critical end node from all the end nodes to lead a conclusion of the entire graph."""
        end_nodes = self.find_end_nodes()
        if not end_nodes:
            return None

        if len(end_nodes) == 1:
            return end_nodes[0]

        edges = [v for k, v in self.edges if isinstance(v, Edge) and v.source in end_nodes]
        critical_edge = max(edges, key=lambda item: item['weight']) if edges else None
        return critical_edge.target if critical_edge else None

    def find_path(self, target: str, source: Optional[str] = None, weight: Optional[Any] = None) -> Any:
        try:
            self._format_Graph()
            return nx.shortest_path(self.graph, source=source, target=target, weight=weight)
        except:
            return None

    def find_all_paths(self, source: str, target: str) -> List[Any]:
        return list(nx.all_simple_paths(self.graph, source=source, target=target))

    def find_critical_path(self) -> tuple[List[Any], int, Dict[str, int]]:
        """
        Finds the critical path in the graph.
        Returns:
            A tuple containing:
                - The critical path (a list of edge identifiers).
                - The duration of the critical path.
                - A dictionary of all paths and their durations.
        """

        all_paths = {}
        for start_node in self.find_start_nodes():
            for end_node in self.find_end_nodes(): # End at nodes with 0 out-degree
                for edge in nx.all_simple_paths(self.graph, source=start_node.identifier, target=end_node.identifier):
                    edge_weight = sum(self.edges.get(item).weight if self.edges.get(item) else 0 for item in edge)
                    all_paths[tuple(edge)] = edge_weight

        if not all_paths:
            return [], 0, all_paths

        critical_path = max(all_paths, key=all_paths.get)
        critical_duration = all_paths[critical_path]

        return list(set(critical_path)), critical_duration, all_paths

    def is_circled(self, node: Node) -> bool:
        """Check if there's a path from the node to itself and return bool."""
        try:
            path = nx.shortest_path(self.graph, source=node, target=node)
            return True if path else False
        except nx.NetworkXNoPath:
            return False


class TaskGraph(Graph):
    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    should_reform: bool = False
    reform_trigger_event: Optional[ReformTriggerEvent] = None
    outputs: Dict[str, TaskOutput] = Field(default_factory=dict, description="stores node identifier and TaskOutput")
    concl_response_schema: Optional[List[ResponseField] | Type[BaseModel]] = Field(default=None, description="stores final response schema in Pydantic class or response fields")
    concl: Optional[TaskOutput] = Field(default=None, description="stores the final or latest conclusion of the entire task graph")
    usage: Optional[UsageMetrics] = None


    def _save(self, title: str, abs_file_path: str = None) -> None:
        """
        Save the graph image in the local directory.
        """
        try:
            import os
            project_root = os.path.abspath(os.getcwd())
            abs_file_path = abs_file_path if abs_file_path else f"{project_root}/.diagrams"
            title = title if title else f"vhq-Diagram-{str(self.id)}"

            os.makedirs(abs_file_path, exist_ok=True)
            plt.savefig(f"{abs_file_path}/{title}.png")

        except Exception as e:
            Logger().log(level="error", message=f"Failed to save the graph {str(self.id)}: {str(e)}", color="red")


    def _handle_human_input(self) -> str | None:
        """Handles input from human."""
        request = None

        print('Proceed? Y/n:')
        x = input()

        if x.lower() == "y":
            Logger().log(message="Ok, proceeding to the next graph execution.", level="info", color="blue")

        else:
            request = input("Request?")
            if request:
                Logger().log(message=f"Ok. regenerating the graph based on your input: ', {request}", level="info", color="blue")
            else:
                Logger().log(message="Cannot recognize your request.", level="error", color="red")
                self.usage = self.usage if self.usage else UsageMetrics(id=self.id)
                self.usage.record_errors(type=ErrorType.HUMAN_INTERACTION)

        return request


    def _handle_usage(self, start_dt: datetime = None, end_dt: datetime = None) -> UsageMetrics:
        usage = self.usage if self.usage else UsageMetrics(id=self.id)

        if self.outputs:
            for item in self.outputs.values():
                if isinstance(item.usage, UsageMetrics):
                    usage = usage.aggregate(metrics=item.usage)

        if start_dt and end_dt:
            usage.record_latency(start_dt, end_dt)

        self.usage = usage
        return usage


    def add_task(self, task: Node | Task) -> Node:
        """Convert `task` to a Node object and add it to G"""

        if isinstance(task, Node) and task.identifier in self.nodes.keys():
            return task

        elif isinstance(task, Task):
            match = []
            for v in self.nodes.values():
                if type(v) == dict and v["node"] and v["node"].task == task:
                    match.append(v["node"])
                elif v.task == task:
                    match.append(v)

            if match:
                return match[0]
            else:
                node = Node(task=task)
                self.add_node(node)
                return node

        else:
            task_node = task if isinstance(task, Node) else Node(task=task)
            self.add_node(task_node)
            return task_node


    def add_dependency(self, source: str, target: str, **edge_attributes) -> None:
        """
        Add an edge that connect task 1 (source) and task 2 (target) using task_node.name as an identifier
        """
        # if not edge_attributes:
        #     Logger().log(level="error", message="Edge attributes are missing.", color="red")
        edge = Edge()
        for k in Edge.model_fields.keys():
            v = edge_attributes.get(k, None)
            if v is not None:
                setattr(edge, k, v)
            else:
                pass
        self.add_edge(source, target, edge)


    def get_task_status(self, identifier: str) -> TaskStatus | None:
        """Retrieves the latest status of the given node"""
        if not identifier or identifier not in self.nodes.keys():
            Logger().log(level="error", message=f"Task node: {identifier} is not in the graph.", color="red")
            return None

        return self._return_node_object(identifier).status


    def visualize(self, layout: str = None, should_save: bool = False):
        from matplotlib.lines import Line2D
        from versionhq.task_graph.colors import white, black, darkgrey, grey, primary, orange, lightgreen, green, darkgreen, darkergreen

        try:
            pos = nx.drawing.nx_agraph.graphviz_layout(self.graph, prog='dot') # 'dot', 'neato', 'fdp', 'sfdp'
        except ImportError:
            pos = nx.spring_layout(self.graph, seed=42) # REFINEME - layout

        node_colors, legend_elements = list(), list()
        for k, v in self.nodes.items():
            status = self.get_task_status(identifier=k)
            node =  v if isinstance(v, Node) else v["node"] if hasattr(v, "node") else None
            if node:
                task_name = node.task.name if node.task and node.task.name else node.task.description if node.task else None
                legend_label = f"ID {str(node.identifier)[0: 6]}...: {task_name}" if task_name else f"ID {str(node.identifier)[0: 6]}..."
                match status:
                    case TaskStatus.NOT_STARTED:
                        node_colors.append(green)
                        legend_elements.append(Line2D([0], [0], marker='o', color='w', label=legend_label, markerfacecolor=green))

                    case TaskStatus.IN_PROGRESS:
                        node_colors.append(darkgreen)
                        legend_elements.append(Line2D([0], [0], marker='o', color='w', label=legend_label, markerfacecolor=darkgreen))

                    case TaskStatus.WAITING:
                        node_colors.append(lightgreen)
                        legend_elements.append(Line2D([0], [0], marker='o', color='w', label=legend_label, markerfacecolor=lightgreen))

                    case TaskStatus.COMPLETED:
                        node_colors.append(darkergreen)
                        legend_elements.append(Line2D([0], [0], marker='o', color='w', label=legend_label, markerfacecolor=darkergreen))

                    case TaskStatus.DELAYED:
                        node_colors.append(orange)
                        legend_elements.append(Line2D([0], [0], marker='o', color='w', label=legend_label, markerfacecolor=orange))

                    case TaskStatus.ON_HOLD:
                        node_colors.append(grey)
                        legend_elements.append(Line2D([0], [0], marker='o', color='w', label=legend_label, markerfacecolor=grey))

                    case TaskStatus.ERROR:
                        node_colors.append(primary)
                        legend_elements.append(Line2D([0], [0], marker='o', color='w', label=legend_label, markerfacecolor=primary))

                    case _:
                        node_colors.append(white)
                        legend_elements.append(Line2D([0], [0], marker='o', color='w', label=legend_label, markerfacecolor=white))

        critical_paths, duration, paths = self.find_critical_path()
        edge_colors = [black if (u, v) in zip(critical_paths, critical_paths[1:]) else darkgrey for u, v in self.graph.edges()]
        edge_widths = [v.weight * 0.5 for v in self.edges.values()]

        # for k, v in self.edges.items():
        #     # edge_weights = nx.get_edge_attributes(self.graph, 'weight')
        #     # edge_colors.append(plt.cm.viridis(v.weight / max(edge_weights.values())))
        #     edge_widths.append(v.weight * 0.5)

        nx.draw(
            self.graph, pos,
            with_labels=True,
            node_size=600,
            node_color=node_colors,
            font_size=8,
            font_color=darkgrey,
            edge_color=edge_colors,
            width=edge_widths,
            arrows=True,
            arrowsize=15,
            arrowstyle='-|>'
        )

        edge_labels = nx.get_edge_attributes(G=self.graph, name="edges")
        nx.draw_networkx_edge_labels(self.graph, pos, edge_labels=edge_labels)

        plt.legend(handles=legend_elements, loc='lower right')
        plt.title(f"vhq-Diagram {str(self.id)}")

        if should_save:
            self._save(title=f"vhq-Diagram {str(self.id)}")
        plt.show(block=False)


    def activate(self, target: Optional[str] = None) -> Tuple[TaskOutput | None, Dict[str, TaskOutput]]:
        """
        Starts to execute all nodes in the graph or a specific node if the target is given, following the given conditons of the edge obeject.
        Then returns tuple of the last task output and all task outputs (self.outputs)
        """

        Logger().log(color="blue", message=f"Start to activate the graph: {str(self.id)}", level="info")
        start_dt = datetime.datetime.now()

        if target:
            if not [k for k in self.nodes.keys() if k == target]:
                Logger().log(level="error", message=f"The node {str(target)} is not in the graph.", color="red")
                return None, None

            # find a shortest path to each in-degree node of the node and see if dependency met.
            node = self._return_node_object(target)
            edges = [v for k, v in self.edges.items() if v.target.identifier == node.identifier]
            res = None

            if not edges or len([item for item in edges if item.dependency_met()]) == len(edges):
                res = node.handle_task_execution()
                self.outputs.update({ target: res })

            return res, self.outputs

        else:
            if not self.edges or not self.nodes:
                Logger().log(level="error", message="TaskGraph needs at least 2 nodes and 1 edge to activate. We'll return None.", color="red")
                return None

            start_nodes = self.find_start_nodes()
            end_nodes = self.find_end_nodes()
            critical_end_node = self.find_critical_end_node()
            critical_path, _, _ = self.find_critical_path()
            res = None

            if end_nodes and len([node for node in end_nodes if node.status == TaskStatus.COMPLETED]) == len(end_nodes):
                res = self.concl if self.concl else critical_end_node.task.output if critical_end_node else  [v.task.output for v in end_nodes.values()][0]

            elif critical_path:
                nodes = [v for k, v in self.nodes.items() if k in critical_path]
                if nodes:
                    for node in nodes:
                        if node.status == TaskStatus.COMPLETED:
                            res = node.task.output

                        else:
                            res = node.handle_task_execution()
                            self.outputs.update({ node.identifier: res })

                            if not res and start_nodes:
                                for node in start_nodes:
                                    res = node.handle_task_execution()
                                    self.outputs.update({ node.identifier: res })
                else:
                    for edge in self.edges.values():
                        res = edge.activate()
                        node_identifier = edge.target.identifier
                        self.outputs.update({ node_identifier: res })

            elif start_nodes:
                for node in start_nodes:
                    res = node.handle_task_execution()
                    self.outputs.update({ node.identifier: res })

            else:
                for k, edge in self.edges.items():
                    res = edge.activate()
                    node_identifier = edge.target.identifier
                    self.outputs.update({ node_identifier: res })

            if self.should_reform:
                target = [k for k in self.outputs.keys()][-1] if self.outputs else self.find_start_nodes()[0].identifier if self.find_start_nodes() else None

                if not target:
                    pass
                else:
                    res, _ = self.handle_reform(target=target)

            self.concl = res
            self.concl_response_schema = self.concl_response_schema if self.concl_response_schema else res.pydantic.__class__ if res.pydantic else None

            end_dt = datetime.datetime.now()
            self._handle_usage(start_dt, end_dt)
            return res, self.outputs


    def evaluate(self, eval_criteria: List[str] = None) -> Evaluation | None:
        """Evaluates the conclusion based on the given eval criteria."""

        if not isinstance(self.concl, TaskOutput):
            return None

        tasks = [v.task for v in self.nodes.values() if v.task and v.task.id == self.concl.task_id]
        task = tasks[0] if tasks else None

        if not task:
            return None

        if not task.eval_criteria:
            task.eval_criteria = eval_criteria

        eval = self.concl.evaluate(task=task)
        return eval


    def handle_reform(self, target: str = None) -> Self:
        task_description = "Improve the given output: "
        if target:
            output = self.outputs[target].raw if self.outputs and target in self.outputs else None
            if output:
                task_description += str(output)

        if self.reform_trigger_event == ReformTriggerEvent.USER_INPUT:
            request = self._handle_human_input()
            task_description += f"You MUST follow the instruction: {request}"

        new_node = Node(task=Task(description=task_description))
        self.add_node(node=new_node)
        self.add_dependency(source=target, target=new_node.identifier)
        return self.activate(target=new_node.identifier)



import json
import numpy as np
from sklearn.metrics import precision_score, recall_score, roc_auc_score, cohen_kappa_score
from typing import List, Tuple, Dict, Any


class LLMJudge:

    class MockLLM:
        def _generate(self, prompt: str) -> str:
            return str(np.random.random())


    def __init__(self, model: MockLLM = None):
        self.model = model if model else self.MockLLM()


    def judge_summary(self, original_text: str, summary: str) -> float:
        prompt = f"""Evaluate the quality of the following summary on a scale of 0 to 1, where 0 is poor and 1 is excellent.
Consider accuracy, completeness, and conciseness.
Original text: {original_text}
Summary: {summary}
Quality score:"""
        response = self.model._generate(prompt)
        score = float(response.strip())
        return score


def generate_summaries(file_path: str, data: List[Dict[str, Any]] = None, summarizer: Any = None) -> List[Tuple[str, str, float]]:
    """Generates a list of tuple with an original text, summary text, and human judge score."""
    if not data:
        with open(file_path, 'r') as file:
            data = json.load(file)
    summaries = []
    for item in data:
        original_text = item['text']
        summary = summarizer.summarize(original_text)
        human_score = item['human_score']
        summaries.append((original_text, summary, human_score))

    return summaries


def validate(judge: LLMJudge, data: List[Tuple[str, str, float]], threshold: float = 0.5):
    human_scores = []
    predicted_scores = []

    for original_text, summary, human_score in data:
        predicted_score = judge.judge_summary(original_text=original_text, summary=summary)
        human_scores.append(human_score)
        predicted_scores.append(predicted_score)

    human_binary = [1 if score >= threshold else 0 for score in human_scores]
    pred_binary = [1 if score >= threshold else 0 for score in predicted_scores]
    precision = precision_score(human_binary, pred_binary, zero_division=0)
    recall = recall_score(human_binary, pred_binary, zero_division=0)
    auroc = roc_auc_score(human_binary, pred_binary, average='weighted')
    kappa = cohen_kappa_score(human_binary, pred_binary)
    return {
        "precision": precision,
        "recall": recall,
        "auroc": auroc,
        "cohen_kappa": kappa
    }



import json
import sqlite3
import datetime
from typing import Any, Dict, List, Optional

from versionhq._utils.logger import Logger
from versionhq.storage.utils import fetch_db_storage_path

storage_path = fetch_db_storage_path()
default_db_name = "task_output"


class TaskOutputSQLiteStorage:
    """
    An SQLite storage class to handle storing task outputs.
    """

    def __init__(self, db_path: str = f"{storage_path}/{default_db_name}.db") -> None:
        self.db_path = db_path
        self._logger = Logger(verbose=True)
        self._initialize_db()


    def _initialize_db(self):
        """
        Initializes the SQLite database and creates LTM table.
        """

        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute(
                    """
                    CREATE TABLE IF NOT EXISTS task_output (
                        task_id TEXT PRIMARY KEY,
                        output JSON,
                        inputs JSON,
                        timestamp DATETIME DEFAULT CURRENT_TIMESTAMP
                    )
                """
                )
                conn.commit()

        except sqlite3.Error as e:
            self._logger.log(level="error", message=f"SQL database initialization failed: {str(e)}", color="red")


    def add(self, task, output: Dict[str, Any], inputs: Dict[str, Any] = {}):
        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute(
                        """INSERT OR REPLACE INTO task_output
                            (task_id, output, inputs, timestamp)
                            VALUES (?, ?, ?, ?)
                        """,
                     (str(task.id), json.dumps(output), json.dumps(inputs), datetime.datetime.now())
                )
                conn.commit()

        except sqlite3.Error as e:
            self._logger.log(level="error", message=f"SAVING TASK OUTPUT ERROR: {e}", color="red")


    def update(self, task_id: str, **kwargs):
        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                fields, values = [], []
                for k, v in kwargs.items():
                    fields.append(f"{k} = ?")
                    values.append(json.dumps(v) if isinstance(v, dict) else v)

                query = f"UPDATE task_output SET {', '.join(fields)} WHERE task = ?"
                values.append(task_id)
                cursor.execute(query, tuple(values))
                conn.commit()

                if cursor.rowcount == 0:
                    self._logger.log(
                        level="warning", message=f"No row found with task_id {task_id}. No update performed.", color="yellow",
                    )

        except sqlite3.Error as e:
            self._logger.log(level="error", message=f"UPDATE TASK OUTPUTS ERROR: {e}", color="red")


    def load(self) -> Optional[List[Dict[str, Any]]]:
        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute("""
                SELECT *
                FROM task_output
                ORDER BY task_id
                """)

                rows = cursor.fetchall()
                results = []
                for row in rows:
                    result = {
                        "task_id": row[0],
                        "output": json.loads(row[1]),
                        "inputs": json.loads(row[2]),
                        "timestamp": row[3],
                    }
                    results.append(result)
                return results

        except sqlite3.Error as e:
            self._logger.log(level="error", message=f"LOADING TASK OUTPUTS ERROR: {e}", color="red")
            return None


    def delete_all(self):
        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute("DELETE FROM task_output")
                conn.commit()

        except sqlite3.Error as e:
            self._logger.log(level="error", message=f"ERROR: Failed to delete all: {e}", color="red")



class TaskOutputStorageHandler:
    """A class to task output storage."""

    from versionhq.task.model import Task

    def __init__(self):
        self.storage = TaskOutputSQLiteStorage()


    def update(self, task: Task, inputs: Dict[str, Any] = {}) -> None:
        saved_outputs = self.load()
        if saved_outputs is None:
            raise ValueError("Logs cannot be None")

        self.add(task, inputs)


    def add(self, task: Task, inputs: Dict[str, Any] = {}) -> None:
        output_to_store = dict(
            id=str(task.id),
            description=str(task.description),
            raw=str(task.output.raw),
            responsible_agents=str(task.processed_agents),
            tokens=task.output.usage.total_tokens,
            latency=task.output.usage.latency,
            score=task.output.aggregate_score if task.output.aggregate_score else "None",
        )
        self.storage.add(task=task, output=output_to_store, inputs=inputs)


    def reset(self) -> None:
        self.storage.delete_all()


    def load(self) -> Optional[List[Dict[str, Any]]]:
        return self.storage.load()



from typing import Any
from pydantic import BaseModel
import versionhq as vhq


class Demo(BaseModel):
    """
    A demo pydantic class to validate the outcome with various nested data types.
    """
    demo_1: int
    demo_2: float
    demo_3: str
    demo_4: bool
    demo_5: list[str]
    demo_6: dict[str, Any]
    demo_nest_1: list[dict[str, Any]]
    demo_nest_2: list[list[str]]
    demo_nest_3: dict[str, list[str]]
    demo_nest_4: dict[str, dict[str, Any]]


demo_response_fields = [
    vhq.ResponseField(title="demo_1", data_type=int),
    vhq.ResponseField(title="demo_2", data_type=float),
    vhq.ResponseField(title="demo_3", data_type=str),
    vhq.ResponseField(title="demo_4", data_type=bool),
    vhq.ResponseField(title="demo_5", data_type=list, items=str),
    vhq.ResponseField(title="demo_6", data_type=dict, properties=[vhq.ResponseField(title="demo-item", data_type=str)]),
    vhq.ResponseField(title="demo_nest_1", data_type=list, items=dict, properties=([
        vhq.ResponseField(title="nest1", data_type=dict, properties=[vhq.ResponseField(title="nest11", data_type=str)])
    ])),
    vhq.ResponseField(title="demo_nest_2", data_type=list, items=list),
    vhq.ResponseField(title="demo_nest_3", data_type=dict, properties=[
        vhq.ResponseField(title="nest1", data_type=list, items=str)
    ]),
    vhq.ResponseField(title="demo_nest_4", data_type=dict, properties=[
        vhq.ResponseField(title="nest1", data_type=dict, properties=[vhq.ResponseField(title="nest12", data_type=str)])
    ])
]




from typing import Dict, List, Tuple, Any
from textwrap import dedent

from pydantic import InstanceOf, BaseModel

from versionhq._utils import is_valid_url, convert_img_url


class Prompt:
    """A class to format, store, and manage a prompt."""

    task: Any = None
    agent: Any = None
    context: Any = None

    def __init__(self, task, agent, context):
        from versionhq.agent.model import Agent
        from versionhq.task.model import Task

        self.task = task if isinstance(task, Task) else Task(description=str(task))
        self.agent = agent if isinstance(agent, Agent) else Agent(role=str(agent))
        self.context = context


    def _draft_output_prompt(self) -> str:
        """Drafts prompt for output format using `response_schema`."""

        from versionhq.task.model import ResponseField

        output_prompt = ""
        output_formats_to_follow = dict()

        if self.task.is_multimodal == False:
            if self.task.response_schema:
                if isinstance(self.task.response_schema, list):
                    for item in self.task.response_schema:
                        if isinstance(item, ResponseField):
                            output_formats_to_follow[item.title] = f"<Return your answer in {item.data_type.__name__}>"

                elif issubclass(self.task.response_schema, BaseModel):
                    for k, v in self.task.response_schema.model_fields.items():
                        output_formats_to_follow[k] = f"<Return your answer in {v.annotation}>"

                output_prompt = f"""Your response MUST be a valid JSON string that strictly follows the response format. Use double quotes for all keys and string values. Do not use single quotes, trailing commas, or any other non-standard JSON syntax.
Ref. Output image: {output_formats_to_follow}"""
            else:
                output_prompt = "You MUST return your response as a valid JSON serializable string, enclosed in double quotes. Use double quotes for all keys and string values. Do NOT use single quotes, trailing commas, or other non-standard JSON syntax."

        else:
            output_prompt = "Return your response in concise manner."
        return dedent(output_prompt)


    def _draft_context_prompt(self, context: Any = None) -> str:
        """
        Create a context prompt from the given context in any format: a task object, task output object, list, dict.
        """
        from versionhq.task.model import Task, TaskOutput

        context_to_add = None
        if not context:
            return context_to_add

        match context:
            case str():
                context_to_add = context

            case Task():
                if not context.output:
                    res = context.execute()
                    context_to_add = res._to_context_prompt()

                else:
                    context_to_add = context.output.raw

            case TaskOutput():
                context_to_add = context._to_context_prompt()


            case dict():
                context_to_add = str(context)

            case list():
                res = ", ".join([self._draft_context_prompt(context=item) for item in context])
                context_to_add = res

            case _:
                pass

        return dedent(context_to_add)


    def _format_content_prompt(self) -> Dict[str, str]:
        """Formats content (file, image, audio) prompt message."""

        import base64
        from pathlib import Path

        content_messages = {}

        if self.task.is_multimodal == False:
            if self.task.image:
                img_url = convert_img_url(self.task.image)
                if img_url:
                    content_messages.update({ "type": "image_url", "image_url": { "url": img_url }})

            if self.task.file:
                if is_valid_url(self.task.file):
                    content_messages.update({ "type": "image_url", "image_url": self.file })

            if self.task.audio and self.agent.llm.provider == "gemini":
                audio_bytes = Path(self.task.audio).read_bytes()
                encoded_data = base64.b64encode(audio_bytes).decode("utf-8")
                content_messages.update({ "type": "image_url", "image_url": "data:audio/mp3;base64,{}".format(encoded_data)})

        return content_messages


    def _find_rag_tools(self) -> List[InstanceOf[Any]]:
        """Find RAG tools from the agent and task object."""

        from versionhq.tool.rag_tool import RagTool

        tools = []
        if self.task.tools:
            [tools.append(item) for item in self.task.tools if isinstance(item, RagTool)]

        if self.agent.tools and self.task.can_use_agent_tools:
            [tools.append(item) for item in self.agent.tools if isinstance(item, RagTool)]

        return tools


    def draft_user_prompt(self) -> str:
        """Draft task prompts from its description and context."""

        output_prompt = self._draft_output_prompt()
        task_slices = [self.task.description, output_prompt, ]

        if self.context:
            context_prompt = self._draft_context_prompt(context=self.context)
            task_slices.insert(len(task_slices), f"Consider the following context when responding: {context_prompt}")

        return "\n".join(task_slices)


    def format_core(self, rag_tools: List[Any] = None, gpt_tools: List[Any] = None) -> Tuple[str, str, List[Dict[str, str]]]:
        """Formats prompt messages sent to the LLM, then returns task prompt, developer prompt, and messages."""

        from versionhq.knowledge._utils import extract_knowledge_context
        from versionhq.memory.contextual_memory import ContextualMemory

        user_prompt = self.draft_user_prompt()
        rag_tools = rag_tools if rag_tools else self._find_rag_tools()

        if self.agent._knowledge:
            agent_knowledge = self.agent._knowledge.query(query=[user_prompt,], limit=5)
            if agent_knowledge:
                agent_knowledge_context = extract_knowledge_context(knowledge_snippets=agent_knowledge)
                if agent_knowledge_context:
                    user_prompt += agent_knowledge_context

        if rag_tools:
            for item in rag_tools:
                rag_tool_context = item.run(agent=self.agent, query=self.task.description)
                if rag_tool_context:
                    user_prompt += ",".join(rag_tool_context) if isinstance(rag_tool_context, list) else str(rag_tool_context)

        if gpt_tools:
            for item in gpt_tools:
                raw, _, _ = item.run()
                if raw:
                    user_prompt += str(raw)

        if self.agent.with_memory == True:
            contextual_memory = ContextualMemory(
                memory_config=self.agent.memory_config, stm=self.agent.short_term_memory, ltm=self.agent.long_term_memory, um=self.agent.user_memory
            )
            context_str = self._draft_context_prompt(context=self.context)
            query = f"{self.task.description} {context_str}".strip()
            memory = contextual_memory.build_context_for_task(query=query)
            if memory.strip() != "":
                user_prompt += memory.strip()


        ## comment out - training
        # if self.agent.networks and self.agent.networks._train:
        #     user_prompt = self.agent._training_handler(user_prompt=user_prompt)
        # else:
        #     user_prompt = self.agent._use_trained_data(user_prompt=user_prompt)

        content_prompt = self._format_content_prompt()

        messages = []
        if content_prompt:
            messages.append(
                {
                    "role": "user",
                    "content": [
                        {
                            "type": "text",
                            "text": user_prompt
                        },
                        content_prompt,
                    ]
                })
        else:
            messages.append({ "role": "user", "content": user_prompt })

        if self.agent.use_developer_prompt:
            messages.append({ "role": "developer", "content": self.agent.backstory })

        return user_prompt, self.agent.backstory if self.agent.use_developer_prompt else None, messages



from enum import Enum

DEFAULT_AUTH_SCHEME = "OAUTH2"

class ComposioAuthScheme(str, Enum):
    OAUTH2 = "OAUTH2"
    BEARER_TOKEN = "BEARER_TOKEN"
    API_KEY = "API_KEY"


class ComposioAppName(str, Enum):
    """
    Enum to store app names that we can connect via Composio as data pipelines or destination services.
    """

    SALESFORCE = "salesforce"
    AIRTABLE = "airtable"
    MAILCHIMP = "mailchimp"
    HUBSPOT = "hubspot"
    KLAVIYO = "klaviyo"
    GOOGLESHEET = "googlesheets"
    GMAIL = "gmail"
    FACEBOOK = "facebook"
    TWITTER = "twitter"
    TWITTER_MEDIA = "twitter_media"
    LINKEDIN = "linkedin"


composio_app_set = [
    (ComposioAppName.SALESFORCE, ComposioAuthScheme.OAUTH2),
    (ComposioAppName.AIRTABLE, ComposioAuthScheme.OAUTH2, ComposioAuthScheme.API_KEY, ComposioAuthScheme.BEARER_TOKEN),
    (ComposioAppName.MAILCHIMP, ComposioAuthScheme.OAUTH2),
    (ComposioAppName.HUBSPOT, ComposioAuthScheme.OAUTH2, ComposioAuthScheme.BEARER_TOKEN),
    (ComposioAppName.KLAVIYO, ComposioAuthScheme.OAUTH2, ComposioAuthScheme.API_KEY),
    (ComposioAppName.GOOGLESHEET, ComposioAuthScheme.OAUTH2),
    (ComposioAppName.GMAIL, ComposioAuthScheme.OAUTH2, ComposioAuthScheme.BEARER_TOKEN),
    (ComposioAppName.TWITTER, ComposioAuthScheme.OAUTH2),
    (ComposioAppName.TWITTER_MEDIA, ComposioAuthScheme.OAUTH2),
    (ComposioAppName.FACEBOOK, ComposioAuthScheme.OAUTH2),
    (ComposioAppName.LINKEDIN, ComposioAuthScheme.OAUTH2),
]

class ComposioStatus(str, Enum):
    INITIATED = "INITIATED"
    ACTIVE = "ACTIVE"
    FAILED = "FAILED"




class ComposioAction(str, Enum):
    """
    Enum to store composio's action that can be called via `Actions.xxx`
    """
    # HUBSPOT_INITIATE_DATA_IMPORT_PROCESS = "hubspot_initate_date_import_process"
    HUBSPOT_CREATE_PIPELINE_STAGE = "hubspot_create_pipeline_stage"



from typing import Type

JSON_URL = "https://raw.githubusercontent.com/BerriAI/litellm/main/model_prices_and_context_window.json"

PROVIDERS = {
    "openai": { "api_key": "OPENAI_API_KEY" },
    "gemini": { "api_key": "GEMINI_API_KEY" },
    "openrouter": { "api_key": "OPENROUTER_API_KEY" },
    "anthropic": { "api_key": "ANTHROPIC_API_KEY" },
    "bedrock":  {
        "aws_access_key_id": "AWS_ACCESS_KEY_ID",
        "aws_secret_access_key": "AWS_SECRET_ACCESS_KEY",
        "aws_region_name": "AWS_REGION_NAME",
    },
    "bedrock/converse":  {
        "aws_access_key_id": "AWS_ACCESS_KEY_ID",
        "aws_secret_access_key": "AWS_SECRET_ACCESS_KEY",
        "aws_region_name": "AWS_REGION_NAME",
    },
    "sagemaker":  {
        "aws_access_key_id": "AWS_ACCESS_KEY_ID",
        "aws_secret_access_key": "AWS_SECRET_ACCESS_KEY",
        "aws_region_name": "AWS_REGION_NAME",
    },
    "huggingface": {
        "api_key": "HUGGINGFACE_API_KEY",
        "base_url": "HUGGINGFACE_API_BASE",
        "HF_ENDPOINT": "HF_ENDPOINT",
    },
    "azure":  {
        "api_base": "AZURE_OPENAI_ENDPOINT_MODEL_NAME",
        "api_key": "AZURE_OPENAI_API_KEY",
        "api_version": "AZURE_OPENAI_API_VERSION",
    },
    "azure_ai": {
        "api_key": "AZURE_AI_API_KEY",
        "base_url": "AZURE_AI_API_BASE",
    }
}

ENDPOINTS = [
    "azure", # endpoints must be aligned with the selected model.
    "azure_ai", # endpoints must be aligned with the selected model.
    "huggingface",
]


# Resaoning and text generation models
MODELS = {
    "openai": [
        "gpt-4.5-preview-2025-02-27",
        "gpt-4",
        "gpt-4o",
        "gpt-4o-mini",
        "o3-mini",
        "o3-mini-2025-01-31",
        "o1-mini",
        "o1-preview",
    ],
    "gemini": [
        "gemini/gemini-2.5-pro-exp-03-25",
        "gemini/gemini-2.0-flash",
        "gemini/gemini-2.0-flash-thinking-exp",
        "gemini/gemini-2.0-flash-lite-preview-02-05",
        "gemini/gemini-2.0-flash-exp",
    ],
    "anthropic": [
        "claude-3-7-sonnet-latest",
        "claude-3-5-haiku-latest",
        "claude-3-5-sonnet-latest",
        "claude-3-opus-latest",
    ],
    "openrouter": [
        "openrouter/deepseek/deepseek-r1",
        "openrouter/qwen/qwen-2.5-72b-instruct",
        "openrouter/google/gemini-2.0-flash-001",
        "openrouter/mistralai/mistral-large",
        "openrouter/cohere/command-r-plus",
        "openrouter/databricks/dbrx-instruct",
    ],
    "bedrock": [
        "bedrock/converse/us.meta.llama3-3-70b-instruct-v1:0",
        "bedrock/us.meta.llama3-2-1b-instruct-v1:0",
        "bedrock/us.meta.llama3-2-3b-instruct-v1:0",
        "bedrock/us.meta.llama3-2-11b-instruct-v1:0",
        "bedrock/us.meta.llama3-2-90b-instruct-v1:0",
        "bedrock/mistral.mistral-7b-instruct-v0:2",
        "bedrock/mistral.mixtral-8x7b-instruct-v0:1",
        "bedrock/mistral.mistral-large-2407-v1:0",
        "bedrock/amazon.titan-text-lite-v1",
        "bedrock/amazon.titan-text-express-v1",
        "bedrock/amazon.titan-text-premier-v1:0",
        "bedrock/cohere.command-r-plus-v1:0",
        "bedrock/cohere.command-r-v1:0",
        "bedrock/cohere.command-text-v14",
        "bedrock/cohere.command-light-text-v14",
    ],
    "azure": [
        "azure/whisper",
        "azure/whisper-2",
        "azure/gpt-4o-mini-audio-preview",

        "azure/DeepSeek-V3",
        "azure/DeepSeek-R1",
        "azure/Llama-3.3-70B-Instruct",
        "azure/Llama-3.2-11B-Vision-Instruct",
        "azure/Meta-Llama-3.1-405B-Instruct",
        "azure/Meta-Llama-3.1-8B-Instruct",
        "azure/Llama-3.2-1B-Instruct",
        "azure/Meta-Llama-3.1-70B",
        "azure/Meta-Llama-3.1-8B",
        "azure/Llama-3.2-3B-Instruct",
        "azure/Meta-Llama-3-8B-Instruct",
        "azure/Meta-Llama-3.1-70B-Instruct",
        "azure/Llama-3.2-90B-Vision-Instruct",
        "azure/Llama-3.2-3B",
        "azure/Llama-3.2-1B",
        "azure/mistral-large-latest",
        "azure/mistral-large-2402",
        "azure/command-r-plus",
        "azure/o3-mini-2025-01-31",
        "azure/o3-mini",
        "azure/o1-mini",
        "azure/Phi-4-mini-instruct",
        "azure/Phi-4-multimodal-instruct",
        "azure/Mistral-Large-2411",
        "azure/Mistral-small"
        "azure/mistral-small-2503",
        "azure/Ministral-3B",
        "azure/mistralai-Mixtral-8x22B-v0-1"
        "azure/Cohere-rerank-v3.5",
    ],
    "azure_ai": [
        "azure_ai/DeepSeek-V3",
        "azure_ai/DeepSeek-R1",
        "azure_ai/Llama-3.3-70B-Instruct",
        "azure_ai/Llama-3.2-11B-Vision-Instruct",
        "azure_ai/Meta-Llama-3.1-405B-Instruct",
        "azure_ai/Meta-Llama-3.1-8B-Instruct",
        "azure_ai/Llama-3.2-1B-Instruct",
        "azure_ai/Meta-Llama-3.1-70B",
        "azure_ai/Meta-Llama-3.1-8B",
        "azure_ai/Llama-3.2-3B-Instruct",
        "azure_ai/Meta-Llama-3-8B-Instruct",
        "azure_ai/Meta-Llama-3.1-70B-Instruct",
        "azure_ai/Llama-3.2-90B-Vision-Instruct",
        "azure_ai/Llama-3.2-3B",
        "azure_ai/Llama-3.2-1B",
        "azure_ai/mistral-large-latest",
        "azure_ai/mistral-large-2402",
        "azure_ai/command-r-plus",
        "azure_ai/o3-mini-2025-01-31",
        "azure_ai/o3-mini",
        "azure_ai/o1-mini",
        "azure_ai/Phi-4-mini-instruct",
        "azure_ai/Phi-4-multimodal-instruct",
        "azure_ai/Mistral-Large-2411",
        "azure_ai/Mistral-small"
        "azure_ai/mistral-small-2503",
        "azure_ai/Ministral-3B",
        "azure_ai/mistralai-Mixtral-8x22B-v0-1"
        "azure_ai/Cohere-rerank-v3.5",
    ],
    "huggingface": [
        "huggingface/qwen/qwen2.5-VL-72B-Instruct",
    ],
}


AUDIO_TO_TEXT_MODELS = [
    "azure/whisper",
    "azure/whisper-2",
    "azure/gpt-4o-mini-audio-preview",
]


"""
Max input token size by the model.
"""
LLM_CONTEXT_WINDOW_SIZES = {
    "gpt-4.5-preview-2025-02-27": 128000,
    "gpt-4o": 128000,
    "gpt-4o-mini": 128000,
    "gpt-4": 8192,
    "o1-preview": 128000,
    "o1-mini": 128000,
    "o3-mini": 200000,
    "o3-mini-2025-01-31": 200000,

    "gemini/gemini-2.0-flash-exp": 1048576,
    "gemini/gemini-2.0-flash": 1048576,
    "gemini/gemini-2.0-flash-thinking-exp": 1048576,
    "gemini/gemini-2.0-flash-lite-preview-02-05": 1048576,

    "claude-3-7-sonnet-latest": 200000,
    "claude-3-5-haiku-latest": 200000,
    "claude-3-5-sonnet-latest": 200000,
    "claude-3-opus-latest": 200000,

    "openrouter/deepseek/deepseek-r1" :65336,
    "openrouter/qwen/qwen-2.5-72b-instruct": 33792,
    "openrouter/google/gemini-2.0-flash-001": 1048576,
    "openrouter/mistralai/mistral-large": 32000,
    "openrouter/cohere/command-r-plus": 128000,
    "openrouter/databricks/dbrx-instruct": 32768,

    "bedrock/converse/us.meta.llama3-3-70b-instruct-v1:0": 8192,
    "bedrock/us.meta.llama3-2-1b-instruct-v1:0": 128000,
    "bedrock/us.meta.llama3-2-3b-instruct-v1:0": 128000,
    "bedrock/us.meta.llama3-2-11b-instruct-v1:0": 128000,
    "bedrock/us.meta.llama3-2-90b-instruct-v1:0": 128000,
    "bedrock/mistral.mistral-7b-instruct-v0:2": 8191,
    "bedrock/mistral.mixtral-8x7b-instruct-v0:1": 8191,
    "bedrock/mistral.mistral-large-2407-v1:0": 8191,
    "bedrock/amazon.titan-text-lite-v1": 4000,
    "bedrock/amazon.titan-text-express-v1": 8000,
    "bedrock/amazon.titan-text-premier-v1:0": 32000,
    "bedrock/cohere.command-r-plus-v1:0": 4096,
    "bedrock/cohere.command-r-v1:0": 4096,
    "bedrock/cohere.command-text-v14": 4096,
    "bedrock/cohere.command-light-text-v14": 4096,
}



"""
Model config related params for litellm.completion().
"""
MODEL_PARAMS = {
    "litellm": [
        "num_retries",
        "context_window_fallback_dict",
        "fallbacks",
        "metadata",
    ],
    "common": [
        # "model",
        # "messages",
        # "stream",
        "response_format",
        "temperature",
        "top_p",
        "max_tokens",
        "tools",
        "tool_choice",
        "n",
        "stop",
    ],
    "openai": [
        "timeout",
        "stream_options",
        "max_compl,etion_tokens",
        "modalities",
        "prediction",
        "audio",
        "presence_penalty",
        "frequency_penalty",
        "logit_bias",
        "user",
        "seed",
        "logprobs",
        "top_logprobs",
        "parallel_tool_calls",
        "extra_headers",
        "model_list"
    ],
    "gemini": [
        "topK",
    ],
    "bedrock": {
        "top-k",
    }
}


class SchemaType:
    """
    A class to store/convert a LLM-valid schema type from the Python Type object.
    https://swagger.io/docs/specification/v3_0/data-models/data-types/
    https://cloud.google.com/vertex-ai/docs/reference/rest/v1/Schema#Type
    """

    def __init__(self, type: Type):
        self.type = type

    def convert(self) -> str:
        if self.type is None:
            return "string"

        if self.type is int:
            return "integer"
        elif self.type is float:
            return "number"
        elif self.type is str:
            return "string"
        elif self.type is dict:
            return "object"
        elif self.type is list:
            return "array"
        elif self.type is bool:
            return "boolean"
        else:
            return "string"



from unittest.mock import patch

from versionhq._utils import UsageMetrics
from versionhq._prompt.model import Prompt


def test_gpt_cua():
    import versionhq as vhq

    params = dict(
        user_prompt="test",
        reasoning_effort="medium",
        browser="firefox",
        tools=vhq.CUAToolSchema(display_width=500, display_height=300, environment="mac", type="computer_use_preview")
    )
    tool = vhq.GPTToolCUA(**params)

    assert tool.user_prompt == params["user_prompt"]
    assert tool.img_url is None
    assert tool.web_url is not None
    assert tool.browser == params["browser"]
    assert tool.reasoning_effort == "medium"
    assert isinstance(tool.tools, list)
    assert tool.tools[0].display_width == 500
    assert tool.tools[0].type == "computer_use_preview"

    with patch.object(vhq.GPTToolCUA, "_structure_schema", return_value=None) as mock_schema:
        tool.run()
        mock_schema.assert_called()


def test_gpt_web_search():
    import versionhq as vhq

    params = dict(input="Search today's top news.", country="US", city="test", search_content_size = "high")
    tool = vhq.GPTToolWebSearch(**params)
    assert tool.model == "gpt-4o"
    assert tool.input == params["input"]
    assert tool._user_location == None
    assert isinstance(tool._usage, UsageMetrics)
    assert tool.search_content_size == "high"
    assert tool.schema is not None

    raw, annotations, usage = tool.run()
    assert raw is not None if usage.total_errors == 0 else raw == ""

    if raw:
        assert isinstance(annotations, list)
        assert isinstance(usage, UsageMetrics)
        assert usage.total_tokens > 0
        assert usage.latency > 0


def test_gpt_file_search():
    import versionhq as vhq

    params = dict(
        input="Search today's top news.",
        vector_store_ids="vs_dummy_id",
        max_num_results=5,
    )
    tool = vhq.GPTToolFileSearch(**params)
    assert tool.model == "gpt-4o"
    assert tool.input == params["input"]
    assert isinstance(tool._usage, UsageMetrics)
    assert tool.vector_store_ids == [params["vector_store_ids"]]
    assert tool.max_num_results == params["max_num_results"]
    assert tool.schema is not None

    with patch.object(vhq.GPTToolFileSearch, "run", return_value=("test", [], UsageMetrics())) as mock_run:
        tool.run()
        mock_run.assert_called_once()


def test_with_agent():
    import versionhq as vhq

    tool_1 = vhq.GPTToolWebSearch(input="Search today's top news.", search_content_size="high")
    tool_2 = vhq.GPTToolFileSearch(input="Search today's top news.", vector_store_ids="vs_dummy_id", max_num_results=5)

    agent = vhq.Agent(role="GPT Tool Handling", tools=[tool_1])
    with patch.object(vhq.Agent, "_handle_gpt_tools", return_value=(vhq.TaskOutput(raw=""))) as mock_run:
        agent.start(tool_res_as_final=True)
        mock_run.assert_called_once()


    mock_value = ("test", "test", [{ "role": "assistant", "content": ""}])
    with patch.object(Prompt, "format_core", return_value=mock_value) as mock_prompt:
        agent.start()
        mock_prompt.assert_called_once()


    agent = vhq.Agent(role="GPT Tool Handling", tools=[tool_2])
    with patch.object(vhq.Agent, "_handle_gpt_tools", return_value=(vhq.TaskOutput(raw=""))) as mock_run:
        agent.start(tool_res_as_final=True)
        mock_run.assert_called_once()

    with patch.object(Prompt, "format_core", return_value=mock_value) as mock_prompt:
        agent.start()
        mock_prompt.assert_called_once()


    agent = vhq.Agent(role="GPT Tool Handling", tools=[tool_1, tool_2])
    with patch.object(vhq.Agent, "_handle_gpt_tools", return_value=("test", "test", vhq.TaskOutput(raw=""), UsageMetrics())) as mock_run:
        agent.start(tool_res_as_final=True)
        mock_run.assert_called_once()

    with patch.object(Prompt, "format_core", return_value=mock_value) as mock_prompt:
        agent.start()
        mock_prompt.assert_called_once()


def test_with_task():
    import versionhq as vhq

    tool_1 = vhq.GPTToolWebSearch(input="Search today's top news.", search_content_size = "high")
    tool_2 = vhq.GPTToolFileSearch(input="Search today's top news.", vector_store_ids="vs_dummy_id", max_num_results=5)

    task = vhq.Task(description="Test gpt tools", tools=[tool_1], tool_res_as_final=True)
    mock_value = ("test", "test", [{ "role": "assistant", "content": ""}])

    with patch.object(vhq.Agent, "_handle_gpt_tools", return_value=(vhq.TaskOutput(raw="", usage=UsageMetrics()))) as mock_run:
        task.execute()
        mock_run.assert_called_once()

    with patch.object(Prompt, "format_core", return_value=mock_value) as mock_prompt:
        task.tool_res_as_final = False
        task.execute()
        mock_prompt.assert_called_once()


    task = vhq.Task(description="Test gpt tools", tools=[tool_2], tool_res_as_final=True)
    with patch.object(vhq.Agent, "_handle_gpt_tools", return_value=(vhq.TaskOutput(raw="", usage=UsageMetrics()))) as mock_run:
        task.execute()
        mock_run.assert_called_once()

    with patch.object(Prompt, "format_core", return_value=mock_value) as mock_prompt:
        task.tool_res_as_final = False
        task.execute()
        mock_prompt.assert_called_once()


    task = vhq.Task(description="Test gpt tools", tools=[tool_1, tool_2], tool_res_as_final=True)
    with patch.object(vhq.Agent, "_handle_gpt_tools", return_value=(vhq.TaskOutput(raw="", usage=UsageMetrics()))) as mock_run:
        task.execute()
        mock_run.assert_called_once()

    with patch.object(Prompt, "format_core", return_value=mock_value) as mock_prompt:
        task.tool_res_as_final = False
        task.execute()
        mock_prompt.assert_called_once()



def test_rag_tool():
    import versionhq as vhq
    rt = vhq.RagTool(url="https://github.com/chroma-core/chroma/issues/3233", query="What is the next action plan?")
    res = rt.run()

    assert rt.text is not None
    assert res is not None


def test_rag_tool_with_agent():
    import versionhq as vhq

    agent = vhq.Agent(role="RAG Tool Tester")
    rt = vhq.RagTool(url="https://github.com/chroma-core/chroma/issues/3233", query="What is the next action plan?")
    res = rt.run(agent=agent)

    assert agent.knowledge_sources is not None
    assert rt.text is not None
    assert res is not None


def test_use_rag_tool():
    import versionhq as vhq

    rt = vhq.RagTool(url="https://github.com/chroma-core/chroma/issues/3233", query="What is the next action plan?")
    agent = vhq.Agent(role="RAG Tool Tester", tools=[rt])
    task = vhq.Task(description="return a simple response", can_use_agent_tools=True, tool_res_as_final=True)
    res = task.execute(agent=agent)

    assert res.final == res.tool_output
    assert res.tool_output is not None



import json
import os
import sys
import threading
import warnings
from dotenv import load_dotenv
from contextlib import contextmanager
from typing import Any, Dict, List, Optional
from typing_extensions import Self

import litellm
from litellm import JSONSchemaValidationError, get_supported_openai_params, supports_response_schema
from pydantic import BaseModel, Field, PrivateAttr, model_validator, ConfigDict

from versionhq.llm.llm_vars import LLM_CONTEXT_WINDOW_SIZES, MODELS, AUDIO_TO_TEXT_MODELS, MODEL_PARAMS, PROVIDERS, ENDPOINTS
from versionhq.tool.model import Tool, ToolSet
from versionhq._utils import Logger, UsageMetrics, ErrorType


load_dotenv(override=True)
LITELLM_API_KEY = os.environ.get("LITELLM_API_KEY")
LITELLM_API_BASE = os.environ.get("LITELLM_API_BASE")
DEFAULT_CONTEXT_WINDOW_SIZE = int(8192 * 0.75)
DEFAULT_MODEL_NAME = os.environ.get("DEFAULT_MODEL_NAME", "gpt-4o-mini")
DEFAULT_MODEL_PROVIDER_NAME = os.environ.get("DEFAULT_MODEL_PROVIDER_NAME", "openai")

class FilteredStream:
    def __init__(self, original_stream):
        self._original_stream = original_stream
        self._lock = threading.Lock()

    def write(self, s) -> int:
        with self._lock:
            if (
                "Give Feedback / Get Help: https://github.com/BerriAI/litellm/issues/new"
                in s
                or "LiteLLM.Info: If you need to debug this error, use `os.environ['LITELLM_LOG'] = 'DEBUG'`"
                in s
            ):
                return 0
            return self._original_stream.write(s)

    def flush(self):
        with self._lock:
            return self._original_stream.flush()


@contextmanager
def suppress_warnings():
    with warnings.catch_warnings():
        litellm.set_verbose = False
        warnings.filterwarnings(action="ignore")
        old_stdout = sys.stdout
        old_stderr = sys.stderr
        sys.stdout = FilteredStream(old_stdout)
        sys.stderr = FilteredStream(old_stderr)

        try:
            yield
        finally:
            sys.stdout = old_stdout
            sys.stderr = old_stderr


class LLM(BaseModel):
    """
    An LLM class to store params to send to the LLM. Use LiteLLM or custom providers for the endpoint.
    """

    _logger: Logger = PrivateAttr(default_factory=lambda: Logger(verbose=True))
    _init_model_name: str = PrivateAttr(default=None)
    _usages: list[Dict[str, int]] = PrivateAttr(default_factory=list)

    model: str = Field(default=None)
    provider: Optional[str] = Field(default=None, description="model provider")
    endpoint_provider: Optional[str] = Field(default=None, description="custom endpoint provider for pass through llm call. require base_url")
    base_url: Optional[str] = Field(default=None, description="api base url for endpoint provider")

    # optional params
    response_format: Optional[Any] = Field(default=None)
    llm_config: Optional[Dict[str, Any]] = Field(default_factory=dict, description="stores valid llm config params")
    callbacks: Optional[List[Any]] = Field(default_factory=list)
    tools: Optional[List[Dict[str, Any]]] = Field(default_factory=list, description="stores a list of tool properties")
    timeout: Optional[float | int] = Field(default=None)
    context_window_size: Optional[int] = Field(default=DEFAULT_CONTEXT_WINDOW_SIZE)

    # LiteLLM specific config
    num_retries: Optional[int] = Field(default=1)
    context_window_fallback_dict: Optional[Dict[str, Any]] = Field(default=None, description="A mapping of model to use if call fails due to context window error")
    fallbacks: Optional[List[Any]]= Field(default=None, description="A list of model names + params to be used, in case the initial call fails")
    metadata: Optional[Dict[str, Any]] = Field(default=None)

    model_config = ConfigDict(extra="allow")

    os.environ['LITELLM_LOG'] = 'DEBUG'


    @model_validator(mode="after")
    def validate_model_and_provider(self) -> Self:
        """
        Validate the given model, provider, interface provider.
        """
        litellm.drop_params = True

        self._init_model_name = self.model

        if self.model is None and self.provider is None:
            self.model = DEFAULT_MODEL_NAME
            self.provider = DEFAULT_MODEL_PROVIDER_NAME

        elif self.model is None and self.provider:
            if self.provider not in PROVIDERS:
                self._logger.log(level="warning", message=f"Invalid model provider is provided. We will assign a default model.", color="yellow")
                self.model = DEFAULT_MODEL_NAME
                self.provider = DEFAULT_MODEL_PROVIDER_NAME

            else:
                provider_model_list = MODELS.get(self.provider)
                if provider_model_list:
                    self.model = provider_model_list[0]
                    self.provider = self.provider
                else:
                    self._logger.log(level="warning", message=f"This provider has not models to be called. We will assign a default model.", color="yellow")
                    self.model = DEFAULT_MODEL_NAME
                    self.provider = DEFAULT_MODEL_PROVIDER_NAME

        elif self.model and self.provider is None:
            model_match = [
                item for item in [
                    [val for val in v if val == self.model][0] for k, v in MODELS.items() if [val for val in v if val == self.model]
                ] if item
            ]
            model_partial_match = [
                item for item in [
                    [val for val in v if val.find(self.model) != -1][0] for k, v in MODELS.items() if [val for val in v if val.find(self.model) != -1]
                ] if item
            ]
            provider_match = [k for k, v in MODELS.items() if k == self.model]

            if model_match:
                self.model = model_match[0]
                self.provider = [k for k, v in MODELS.items() if self.model in v][0]

            elif model_partial_match:
                self.model = model_partial_match[0]
                self.provider = [k for k, v in MODELS.items() if [item for item in v if item.find(self.model) != -1]][0]

            elif provider_match:
                provider = provider_match[0]
                if self.MODELS.get(provider):
                    self.provider = provider
                    self.model = self.MODELS.get(provider)[0]
                else:
                    self.provider = DEFAULT_MODEL_PROVIDER_NAME
                    self.model = DEFAULT_MODEL_NAME

            else:
                self.model = DEFAULT_MODEL_NAME
                self.provider = DEFAULT_MODEL_PROVIDER_NAME

        else:
            provider_model_list = MODELS.get(self.provider)
            if self.model not in provider_model_list:
                self._logger.log(level="warning", message=f"The provided model: {self._init_model_name} is not in the list. We will assign a default model.", color="yellow")
                self.model = DEFAULT_MODEL_NAME
                self.provider = DEFAULT_MODEL_PROVIDER_NAME

        # trigger passing through litellm and use original endpoint.
        if self.provider in ENDPOINTS:
            self.endpoint_provider = self.provider

        return self


    @model_validator(mode="after")
    def setup_config(self) -> Self:
        """
        Set up valid config params after setting up a valid model, provider, interface provider names.
        """
        litellm.drop_params = True

        self._set_credentials()

        if self.callbacks:
            self._set_callbacks(self.callbacks)

        self.context_window_size = self._get_context_window_size()

        if self.llm_config:
            self._create_valid_params(config=self.llm_config)

        return self


    def _create_valid_params(self, config: Dict[str, Any]) -> Dict[str, Any]:
        """
        Returns valid params incl. model + litellm original params) from the given config dict.
        """
        valid_keys = list()

        if self.model:
            valid_keys = get_supported_openai_params(model=self.model, custom_llm_provider=self.endpoint_provider, request_type="chat_completion")

        if not valid_keys:
            valid_keys = MODEL_PARAMS.get("common")

        valid_keys += MODEL_PARAMS.get("litellm")

        valid_config = dict()
        for item in valid_keys:
            if hasattr(self, item) and (getattr(self, item) or getattr(self, item) == False):
                valid_config[item] = getattr(self, item)
            elif item in self.llm_config and (self.llm_config[item] or self.llm_config[item]==False):
                valid_config[item] = self.llm_config[item]
            elif item in config and (config[item] or config[item] == False):
                valid_config[item] = config[item]
            else:
                pass

        if "max_tokens" in valid_config:
            if self.context_window_size and valid_config["max_tokens"] > self.context_window_size:
                valid_config["max_tokens"] = self.context_window_size

        self.llm_config = valid_config
        return valid_config


    def _set_credentials(self) -> Dict[str, Any]:
        cred = PROVIDERS.get(self.provider, None) if self.provider else None
        if not cred:
            return {}

        valid_cred = {}
        for k, v in cred.items():
            val = None
            if '_MODEL_NAME' in v:
                model_name = self.model.split('/')[-1] if self.model.split('/') else self.model
                key = v.replace('_MODEL_NAME', f'_{model_name.replace("-", '_').replace(' ', '_').upper()}')
                val = os.environ.get(key, None)
                if not val:
                    val = os.environ.get(v.replace('_MODEL_NAME', ''), None)
            else:
                val = os.environ.get(v, None)

            if val:
                valid_cred[str(k)] = val

        return valid_cred


    def _supports_response_schema(self) -> bool:
        try:
            return supports_response_schema(model=self.model, custom_llm_provider=self.endpoint_provider)
        except:
            return False


    def _supports_function_calling(self) -> bool:
        try:
            if self.model:
                params = get_supported_openai_params(model=self.model)
                return "response_format" in params if params else False
            else:
                return False
        except Exception as e:
            self._logger.log(level="warning", message=f"Failed to get supported params: {str(e)}", color="yellow")
            return False


    def _supports_stop_words(self) -> bool:
        supported_params = get_supported_openai_params(model=self.model, custom_llm_provider=self.endpoint_provider)
        return "stop" in supported_params if supported_params else False


    def _get_context_window_size(self) -> int:
        """
        Only use 75% of the context window size to avoid cutting the message in the middle.
        """
        return int(LLM_CONTEXT_WINDOW_SIZES.get(self.model) * 0.75) if LLM_CONTEXT_WINDOW_SIZES.get(self.model) is not None else DEFAULT_CONTEXT_WINDOW_SIZE


    def _set_callbacks(self, callbacks: List[Any]):
        callback_types = [type(callback) for callback in callbacks]
        for callback in litellm.success_callback[:]:
            if type(callback) in callback_types:
                litellm.success_callback.remove(callback)

        for callback in litellm._async_success_callback[:]:
            if type(callback) in callback_types:
                litellm._async_success_callback.remove(callback)

        litellm.callbacks = callbacks


    def call(
        self,
        messages: List[Dict[str, str]],
        response_format: Optional[Dict[str, Any]] = None,
        tools: Optional[List[Tool | ToolSet | Any ]] = None,
        config: Optional[Dict[str, Any]] = dict(),
        tool_res_as_final: bool = False,
        file: str = None
    ) -> str:
        """Configures and calls the LLM (chat, text generation, reasoning models)."""

        litellm.drop_params = True
        litellm.set_verbose = True

        with suppress_warnings():
            if len(self.callbacks) > 0:
                self._set_callbacks(self.callbacks)

            try:
                res = None
                tool_res = ""
                cred = self._set_credentials()

                if file and self.model in AUDIO_TO_TEXT_MODELS:
                    params = self._create_valid_params(config=config)
                    audio_file = open(file, 'rb')
                    res = litellm.transcription(
                        model=self.model,
                        file=audio_file,
                        rompt=messages,
                        ustom_llm_provider=self.endpoint_provider,
                        response_format="json",
                        **cred
                    )
                    usage = UsageMetrics()
                    if res:
                        usage.latency = res._response_ms if hasattr(res, '_response_ms') else 0
                        self._usages.append(usage)
                        return res.text
                    else:
                        usage.record_errors(type=ErrorType.API)
                        self._usages.append(usage)
                        return None


                if self.provider == "gemini":
                    self.response_format = { "type": "json_object" } if not tools and self.model != "gemini/gemini-2.0-flash-thinking-exp" else None
                elif response_format and "json_schema" in response_format:
                    self.response_format = response_format  if self._supports_function_calling() else  { "type": "json_object" }
                else:
                    self.response_format = response_format

                if not tools:
                    params = self._create_valid_params(config=config)
                    res = litellm.completion(model=self.model, messages=messages, stream=False, **params, **cred)
                    self._usages.append(res["usage"])
                    return res["choices"][0]["message"]["content"]

                else:
                    try:
                        self.tools = [item.tool.properties if isinstance(item, ToolSet) else item.properties for item in tools]
                        params = self._create_valid_params(config=config)
                        res = litellm.completion(model=self.model, messages=messages, **params, **cred)
                        tool_calls = res.choices[0].message.tool_calls

                        if tool_calls:
                            for item in tool_calls:
                                func_name = item.function.name
                                func_args = item.function.arguments
                                if not isinstance(func_args, dict):
                                    try:
                                        func_args = json.loads(json.dumps(eval(str(func_args))))
                                    except:
                                        pass

                                # find a tool whose name is matched with the retrieved func_name
                                matches = []
                                for tool in tools:
                                    tool_name = tool.tool.name if isinstance(tool, ToolSet) else tool.name
                                    tool_func_name = tool.tool.func.__name__ if isinstance(tool, ToolSet) else tool.func.__name__
                                    if tool_name.replace(" ", "_") == func_name or tool_func_name == func_name or tool_name == "random_func":
                                        matches.append(tool)
                                    else:
                                        pass

                                if matches:
                                    tool_to_execute = matches[0]
                                    tool_instance = tool_to_execute.tool if isinstance(tool_to_execute, ToolSet) else tool_to_execute
                                    params = tool_to_execute.kwargs if isinstance(tool_to_execute, ToolSet) else func_args
                                    tool_res_to_add = tool_instance.run(params=params) if params else tool_instance.run()

                                    if tool_res_as_final:
                                        if tool_res_to_add not in tool_res:
                                            tool_res += str(tool_res_to_add)
                                    else:
                                        messages.append(res.choices[0].message)
                                        messages.append({ "role": "tool", "tool_call_id": item.id, "content": str(tool_res_to_add) })

                        else:
                            if tool_res_as_final and tools and not tool_res:
                                for item in tools:
                                    tool_res_to_add = item.tool.run(params=item.kwargs) if isinstance(item, ToolSet) else item.run()
                                    if tool_res_to_add not in tool_res:
                                        tool_res += str(tool_res_to_add)
                                    else:
                                        pass

                    except:
                        if tool_res_as_final and tools and not tool_res:
                            for item in tools:
                                tool_res_to_add = item.tool.run(params=item.kwargs) if isinstance(item, ToolSet) else item.run()
                                if tool_res_to_add not in tool_res:
                                    tool_res += str(tool_res_to_add)
                                else:
                                    pass
                        elif tools and not tool_res:
                                tool_res = res["choices"][0]["message"]["content"]
                        else:
                            pass

                if tool_res_as_final:
                    return tool_res
                else:
                    res = litellm.completion(model=self.model, messages=messages, **params, **cred)
                    self._usages.append(res["usage"])
                    return res.choices[0].message.content

            except JSONSchemaValidationError as e:
                self._logger.log(level="error", message="Raw Response: {}".format(e.raw_response), color="red")
                raise e

            except Exception as e:
                self._logger.log(level="error", message=f"{self.model} failed to execute: {str(e)}", color="red")
                if "litellm.RateLimitError" in str(e):
                    raise e



from typing import List, Dict, Any


def extract_knowledge_context(knowledge_snippets: List[Dict[str, Any]]) -> str:
    """
    Extract knowledge from the task prompt.
    """

    valid_snippets = [result["context"] for result in knowledge_snippets if result and result.get("context")]
    snippet = "\n".join(valid_snippets)
    return f"Additional Information: {snippet}" if valid_snippets else ""



"""
Test use cases on index.md and quickstart.md
"""

def test_doc_index_a():
    import versionhq as vhq

    network = vhq.form_agent_network(
        task="YOUR AMAZING TASK OVERVIEW",
        expected_outcome="YOUR OUTCOME EXPECTATION",
    )
    res, _ = network.launch()
    assert res.final is not None


def test_doc_index_b():
    import versionhq as vhq
    from pydantic import BaseModel

    class CustomOutput(BaseModel):
        test1: str
        test2: list[str]

    def dummy_func(message: str, **kwargs) -> str:
        test1 = kwargs["test1"] if kwargs and "test1" in kwargs else ""
        test2 = kwargs["test2"] if kwargs and "test2" in kwargs else ""
        if test1 and test2:
            return f"""{message}: {test1}, {", ".join(test2)}"""

    agent = vhq.Agent(role="demo", maxit=1)

    task = vhq.Task(
        description="Amazing task",
        response_schema=CustomOutput,
        callback=dummy_func,
        callback_kwargs=dict(message="Hi! Here is the result: ")
    )

    res = task.execute(agent=agent, context="amazing context to consider.")
    assert "Hi! Here is the result:" in res.callback_output
    assert [getattr(res.pydantic, k) for k in CustomOutput.model_fields.keys()]


def test_doc_index_c():
    import versionhq as vhq

    agent_a = vhq.Agent(role="agent a", llm="llm-of-your-choice")
    agent_b = vhq.Agent(role="agent b", llm="llm-of-your-choice")

    task_1 = vhq.Task(
        description="Analyze the client's business model.",
        response_schema=[vhq.ResponseField(title="test1", data_type=str, required=True),],
        allow_delegation=True
    )

    task_2 = vhq.Task(
        description="Define a cohort.",
        response_schema=[vhq.ResponseField(title="test1", data_type=int, required=True),],
        allow_delegation=False
    )

    network =vhq.AgentNetwork(
        members=[
            vhq.Member(agent=agent_a, is_manager=False, tasks=[task_1]),
            vhq.Member(agent=agent_b, is_manager=True, tasks=[task_2]),
        ],
    )
    res, tg = network.launch()

    assert isinstance(res, vhq.TaskOutput)
    assert isinstance(tg, vhq.TaskGraph)
    assert agent_b.key in task_1.processed_agents # delegated by agent_a
    assert agent_b.key in task_2.processed_agents



import pytest
from typing import Any, Callable

from versionhq.agent.model import Agent
from versionhq.llm.model import LLM
from versionhq.task.model import Task
from versionhq.tool.model import Tool

from tests import Demo, demo_response_fields

"""
Test connection to the llm via endpoint provider or litellm.
"""

def set_agent(llm: str, **kwargs) -> Agent:
    agent = Agent(role="LLM Connection Tester", llm=llm, maxit=1, max_retry_limit=1, llm_config=dict(max_tokens=3000), **kwargs)
    return agent

@pytest.fixture(scope='module')
def simple_task():
    return Task(description="write a random poem.")


@pytest.fixture(scope='module')
def tool_task():
    class DemoTool(Tool):
        func: Callable[..., Any] = lambda x: "Demo"
    return Task(description="Simply execute the given tools.", tools=[DemoTool,], tool_res_as_final=True)


@pytest.fixture(scope='module')
def schema_task():
    return Task(description="Generate a random value that adheres strictly to the specified data type.", response_schema=Demo)


@pytest.fixture(scope='module')
def res_field_task():
    return Task(description="return random values strictly following the given response format.", response_schema=demo_response_fields)


def _test_con_bedrock(simple_task, tool_task, schema_task, res_field_task):
    llms_to_test = [
        "bedrock/converse/us.meta.llama3-3-70b-instruct-v1:0",
        "bedrock/us.meta.llama3-2-11b-instruct-v1:0",
        "bedrock/mistral.mistral-7b-instruct-v0:2",
        "bedrock/amazon.titan-text-lite-v1",
    ]
    agents = [set_agent(llm=llm) for llm in llms_to_test]

    for agent in agents:
        assert isinstance(agent.llm, LLM)
        assert agent.llm.provider == "bedrock"
        assert agent.llm._init_model_name and agent.llm.provider and agent.llm.llm_config["max_tokens"] == agent.llm_config["max_tokens"]

        res_1 = simple_task.execute(agent=agent, context="running a test")
        assert res_1.raw is not None

        res_2 = tool_task.execute(agent=agent, context="running a test")
        assert res_2.tool_output is not None

        res_3 = schema_task.execute(agent=agent, context="running a test")
        assert [
            getattr(res_3.pydantic, k) and v.annotation == Demo.model_fields[k].annotation
            for k, v in res_3.pydantic.model_fields.items()
        ]

        res_4 = res_field_task.execute(agent=agent, context="running a test")
        assert [v and type(v) == res_field_task.response_schema[i].data_type for i, (k, v) in enumerate(res_4.json_dict.items())]


def _test_con_openai(simple_task, tool_task, schema_task, res_field_task):
    llms_to_test = [
        "gpt-4.5-preview-2025-02-27",
        "o3-mini",
        "o3-mini-2025-01-31",
    ]
    agents = [set_agent(llm=llm, provider="openai") for llm in llms_to_test]

    for agent in agents:
        assert isinstance(agent.llm, LLM)
        assert agent.llm.provider == "openai"
        assert agent.llm._init_model_name and agent.llm.provider and agent.llm.llm_config["max_tokens"] == agent.llm_config["max_tokens"]

        res_1 = simple_task.execute(agent=agent, context="running a test")
        assert res_1.raw is not None

        res_2 = tool_task.execute(agent=agent, context="running a test")
        assert res_2.tool_output is not None

        res_3 = schema_task.execute(agent=agent, context="running a test")
        assert [
            getattr(res_3.pydantic, k) and v.annotation == Demo.model_fields[k].annotation
            for k, v in res_3.pydantic.model_fields.items()
        ]

        res_4 = res_field_task.execute(agent=agent, context="running a test")
        assert [v and type(v) == res_field_task.response_schema[i].data_type for i, (k, v) in enumerate(res_4.json_dict.items())]


def _test_con_gemini(simple_task, tool_task, schema_task, res_field_task):
    llms_to_test = [
        "gemini/gemini-2.5-pro-exp-03-25"
        "gemini/gemini-2.0-flash",
        "gemini/gemini-2.0-flash-thinking-exp",
        "gemini/gemini-2.0-flash-lite-preview-02-05",
        "gemini/gemini-2.0-flash-exp",
    ]

    agents = [set_agent(llm=llm) for llm in llms_to_test]

    for agent in agents:
        assert isinstance(agent.llm, LLM)
        assert agent.llm.provider == "gemini"
        assert agent.llm._init_model_name and agent.llm.provider and agent.llm.llm_config["max_tokens"] == agent.llm_config["max_tokens"]

        res_1 = simple_task.execute(agent=agent, context="running a test")
        assert res_1.raw is not None

        res_2 = tool_task.execute(agent=agent, context="running a test")
        assert res_2.tool_output is not None

        schema_task = Task(description="Generate a random value that adheres strictly to the specified data type.", response_schema=Demo)
        res_3 = schema_task.execute(agent=agent, context="running a test")
        assert [
            getattr(res_3.pydantic, k) and v.annotation == Demo.model_fields[k].annotation
            for k, v in res_3.pydantic.model_fields.items()
        ]

        res_4 = res_field_task.execute(agent=agent, context="running a test")
        assert [v and type(v) == res_field_task.response_schema[i].data_type for i, (k, v) in enumerate(res_4.json_dict.items())]


def _test_con_azure(simple_task, tool_task, schema_task, res_field_task):
    llms_to_test = [
        "azure/Phi-4-mini-instruct",
        "azure_ai/Phi-4-mini-instruct",
        "azure_ai/DeepSeek-V3",
    ]
    agents = [set_agent(llm=llm) for llm in llms_to_test]

    for agent in agents:
        assert isinstance(agent.llm, LLM)
        assert "azure" in agent.llm.provider
        assert agent.llm._init_model_name and agent.llm.provider and agent.llm.llm_config["max_tokens"] == agent.llm_config["max_tokens"]

        res_1 = simple_task.execute(agent=agent, context="running a test")
        assert res_1.raw is not None

        res_2 = tool_task.execute(agent=agent, context="running a test")
        assert res_2.tool_output is not None

        res_3 = schema_task.execute(agent=agent, context="running a test")
        if res_3.pydantic:
            assert [
                getattr(res_3.pydantic, k) and v.annotation == Demo.model_fields[k].annotation
                for k, v in res_3.pydantic.model_fields.items()
            ]

        res_4 = res_field_task.execute(agent=agent, context="running a test")
        if res_4.json_dict:
            assert [v and type(v) == res_field_task.response_schema[i].data_type for i, (k, v) in enumerate(res_4.json_dict.items())]



from typing import List, Type
from enum import Enum

from pydantic import BaseModel, create_model, Field

from versionhq.task.model import Task
from versionhq.agent.model import Agent
from versionhq.agent_network.model import AgentNetwork, Member, Formation
from versionhq.agent.inhouse_agents  import vhq_formation_planner
from versionhq._utils import Logger, is_valid_enum

import chromadb
chromadb.api.client.SharedSystemClient.clear_system_cache()


def form_agent_network(
        task: str,
        expected_outcome: str | Type[BaseModel],
        agents: List[Agent] = None,
        context: str = None,
        formation: Type[Formation] = None
    ) -> AgentNetwork | None:
    """
    Make a formation of agents from the given task description, expected outcome, agents (optional), and context (optional).
    """

    if not task:
        Logger(verbose=True).log(level="error", message="Missing task description.", color="red")
        return None

    if not expected_outcome:
        Logger(verbose=True).log(level="error", message="Missing expected outcome.", color="red")
        return None

    if formation:
        try:
            match formation:
                case Formation():
                    pass

                case str():
                    matched = [item for item in Formation.s_ if item == formation.upper()]
                    if matched:
                        formation = getattr(Formation, matched[0])
                    else:
                        # Formation._generate_next_value_(name=f"CUSTOM_{formation.upper()}", start=100, count=6, last_values=Formation.HYBRID.name)
                        Logger(verbose=True).log(level="warning", message=f"The formation {formation} is invalid. We'll recreate a valid formation.", color="yellow")
                        formation = None

                case int():
                    formation = Formation(int(formation))

                case float():
                    formation = Formation(round(formation))

                case _:
                    Logger(verbose=True).log(level="warning", message=f"The formation {formation} is invalid. We'll recreate a valid formation.", color="yellow")
                    formation = None

        except Exception as e:
            Logger(verbose=True).log(level="warning", message=f"The formation {formation} is invalid: {str(e)}. We'll recreate a formation.", color="yellow")
            formation = None

    # try:
    prompt_formation = formation.name if formation and isinstance(formation, Formation) else f"Select the best formation to effectively execute the tasks from the given Enum sets: {str(Formation.__dict__)}."

    prompt_expected_outcome = expected_outcome if isinstance(expected_outcome, str) else str(expected_outcome.model_dump()) if type(expected_outcome) == BaseModel else ""

    class Outcome(BaseModel):
        formation: str
        agent_roles: list[str]
        task_descriptions: list[str]
        task_outcomes: list[list[str]]
        leader_agent: str

    vhq_task = Task(
        description=f"Design a team of specialized agents to fully automate the following task and deliver the expected outcome. For each agent, define its role, task description, and expected outputs via the task with items in a list. Then specify the formation if the formation is not given. If you think SUPERVISING or HYBRID is the best formation, include a leader_agent role, else leave the leader_agent role blank.\nTask: {str(task)}\nExpected outcome: {prompt_expected_outcome}\nFormation: {prompt_formation}",
        response_schema=Outcome
    )

    if agents:
        vhq_task.description += "You MUST add the following agents' roles in the network formation: " + ", ".join([agent.role for agent in agents if isinstance(agent, Agent)])

    res = vhq_task.execute(agent=vhq_formation_planner, context=context)

    network_tasks = []
    members = []
    leader = res._fetch_value_of(key="leader_agent")
    agent_roles =  res._fetch_value_of(key="agent_roles")
    created_agents = [Agent(role=str(item), goal=str(item)) for item in agent_roles] if agent_roles else []
    task_descriptions = res._fetch_value_of(key="task_descriptions")
    task_outcomes = res._fetch_value_of(key="task_outcomes")
    formation_key = res.json_dict["formation"] if "formation" in res.json_dict else None
    _formation = Formation[formation_key] if is_valid_enum(key=formation_key, enum=Formation) else Formation.SUPERVISING

    if agents:
        for i, agent in enumerate(created_agents):
            matches = [item for item in agents if item.role == agent.role]
            if matches:
                created_agents[i] = matches[0]
            else:
                pass

    created_tasks = []

    if task_outcomes:
        for i, item in enumerate(task_outcomes):
            if len(task_descriptions) > i and task_descriptions[i]:
                fields = {}
                for ob in item:
                    try:
                        field_name = str(ob).lower().split(":")[0].replace(" ", "_")[0: 16]
                        fields[field_name] = (str, Field(default=None))
                    except:
                        pass
                output = create_model("Output", **fields) if fields else None
                _task = Task(description=task_descriptions[i], response_schema=output)
                created_tasks.append(_task)

    if len(created_tasks) <= len(created_agents):
        for i in range(len(created_tasks)):
            is_manager = False if not leader else bool(created_agents[i].role.lower() == leader.lower())
            member = Member(agent=created_agents[i], is_manager=is_manager, tasks=[created_tasks[i]])
            members.append(member)

        for i in range(len(created_tasks), len(created_agents)):
            try:
                is_manager = False if not leader else bool(created_agents[i].role.lower() == leader.lower())
                member_w_o_task = Member(agent=created_agents[i], is_manager=is_manager)
                members.append(member_w_o_task)
            except:
                pass

    elif len(created_tasks) > len(created_agents):
        for i in range(len(created_agents)):
            is_manager = False if not leader else bool(created_agents[i].role.lower() == leader.lower())
            member = Member(agent=created_agents[i], is_manager=is_manager, tasks=[created_tasks[i]])
            members.append(member)

        network_tasks.extend(created_tasks[len(created_agents):len(created_tasks)])

    if _formation == Formation.SUPERVISING and not [member for member in members if member.is_manager]:
        role = leader if leader else "Leader"
        manager = Member(agent=Agent(role=role), is_manager=True)
        members.append(manager)

    members.sort(key=lambda x: x.is_manager == False)
    network = AgentNetwork(members=members, formation=_formation, network_tasks=network_tasks)

    Logger().log(level="info", message=f"Successfully created a agent network: {str(network.id)} with {len(network.members)} agents.", color="blue")

    return network


    # except Exception as e:
    #     Logger().log(level="error", message=f"Failed to create a agent network - return None. You can try with solo agent. Error: {str(e)}", color="red")
    #     return None



from pathlib import Path
from typing import List
from unittest.mock import patch

import pytest

from versionhq import (
    StringKnowledgeSource,
    TextFileKnowledgeSource,
    PDFKnowledgeSource,
    CSVKnowledgeSource,
    JSONKnowledgeSource,
    ExcelKnowledgeSource,
    DoclingSource
)

str_content = "Kuriko's favorite color is black, and she likes Japanese food."
str_content_long = (
    "Kuriko drinks her coffee black every morning. It is her morning ritual."
    "Kuriko designs bridges."
    "Kuriko owns a fluffy cat. Her name is Luna."
    "Kuriko enjoys solving complex problems."
    "Kuriko's cat naps on her keyboard."
    "Kuriko loves the smell of freshly brewed coffee."
    "Kuriko works on a computer all day and her eyesights are getting poor."
    "Kuriko's apartment is filled with cat toys. She loves to spoil her cat."
    "Kuriko reads engineering journals."
    "Kuriko enjoys a peaceful morning at home."
    "Kuriko is a skilled problem-solver and good at math."
    "Luna loves to play fetch and naps on Kuriko's keyboard."
    "Kuriko works on building projects."
    "Kuriko enjoys trying new coffee blends."
    "Kuriko uses CAD software."
    "Kuriko recently bought 3 Android smartphones."
    "Kurko's cat is her best friend."
    "Kuriko works in a busy office with 100 people on the floor."
    "And she works long hours."
    "Luna is 10 months old."
    "She speaks 3 languages."
    "She is passionate about weight training, tracking weekly volume load to grow muscle."
    "Kuriko recently traveled across Latin America."
    "random sentence, random sentence, random sentence."
)
str_content_long_2 = (
    "J drinks his coffee black every morning."
    "J designs bridges."
    "J owns a fluffy cat. Her name is Luna."
    "J enjoys solving complex problems."
    "J's cat naps on his keyboard."
    "J loves the smell of freshly brewed coffee."
    "J works on a computer all day."
    "J reads engineering journals."
    "J is a skilled problem-solver and good at math."
    "Luna loves to play fetch and naps on J's keyboard."
    "J works on building projects."
    "J enjoys trying new coffee blends."
    "J uses CAD software."
    "J recently bought a laptop."
    "J works in a busy office with 100 people on the floor."
    "And he works long hours."
    "Luna is 10 months old."
    "J recently traveled across Southeast Asia."
)


@pytest.fixture(autouse=True)
def mock_vector_db():
    """
    Mock vector database operations. the query method to return a predefined response
    """

    with patch("versionhq.knowledge.storage.KnowledgeStorage") as mock:
        instance = mock.return_value
        instance.query.return_value = [
            {
                "context": "Kuriko's favorite color is black, and she likes Japanese food.",
                "score": 0.9,
            }
        ]
        instance.reset.return_value = None
        yield instance


@pytest.fixture(autouse=True)
def reset_knowledge_storage(mock_vector_db):
    """
    Fixture to reset knowledge storage before each test.
    """
    yield


def test_string_knowledge_source(mock_vector_db):
    """
    Create a plane text storage source and query them.
    """

    string_source = StringKnowledgeSource(content=str_content, metadata={ "preference": "personal" })
    mock_vector_db.sources = [string_source]
    mock_vector_db.query.return_value = [{ "context": str_content, "score": 0.9 }]

    query = "What is Kuriko's favorite color?"
    results = mock_vector_db.query(query)

    assert any("black" in result["context"].lower() for result in results)
    mock_vector_db.query.assert_called_once()


# @pytest.mark.vcr(filter_headers=["authorization"])
def test_2k_string_knowledge_source(mock_vector_db):
    string_source = StringKnowledgeSource(content=str_content_long, metadata={ "preference": "personal" })
    mock_vector_db.sources = [string_source]
    mock_vector_db.query.return_value = [{ "context": str_content_long, "score": 0.9 }]

    query = "What is Kuriko's cat name?"
    results = mock_vector_db.query(query)

    assert any("luna" in result["context"].lower() for result in results)
    mock_vector_db.query.assert_called_once()


def test_multiple_string_knowledge_sources(mock_vector_db):
    contents = [
        "Kuriko decided to take a stray kitten home.",
        "She enjoys quiet and peaceful morning.",
        "She named the kitten Luna."
    ]
    string_sources = [StringKnowledgeSource(content=content, metadata={ "preference": "personal" }) for content in contents]
    mock_vector_db.query.return_value = [{ "context": "Kuriko has a cat named Luna.", "score": 0.9 }]
    mock_vector_db.sources = string_sources

    query = "What is her cat's name?"
    results = mock_vector_db.query(query)

    assert any("luna" in result["context"].lower() for result in results)
    mock_vector_db.query.assert_called_once()


def test_multiple_2k_string_knowledge_sources(mock_vector_db):
    contents = [str_content_long * 2, str_content_long_2 * 2]
    string_sources = [StringKnowledgeSource(content=content, metadata={ "preference": "personal" }) for content in contents]
    mock_vector_db.sources = string_sources
    mock_vector_db.query.return_value = [{ "context": contents[1], "score": 0.9 }]

    query = "What did J buy recently?"
    results = mock_vector_db.query(query)

    assert any("laptop" in result["context"].lower() for result in results)
    mock_vector_db.query.assert_called_once()


def test_single_short_file_knowledge_source(mock_vector_db, tmpdir):
    """
    Create a text file with a short content and query
    """

    file_path = Path(tmpdir.join("short_file.txt"))
    with open(file_path, "w") as f:
        f.write(str_content)

    file_source = TextFileKnowledgeSource(file_paths=[file_path], metadata={ "preference": "personal" })
    mock_vector_db.sources = [file_source]
    mock_vector_db.query.return_value = [{"context": str_content, "score": 0.9}]

    query = "What color does Kuriko like?"
    results = mock_vector_db.query(query)

    assert any("black" in result["context"].lower() for result in results)
    mock_vector_db.query.assert_called_once()


def test_single_2k_character_file(mock_vector_db, tmpdir):
    file_path = Path(tmpdir.join("long_file.txt"))
    with open(file_path, "w") as f:
        f.write(str_content_long)

    file_source = TextFileKnowledgeSource(file_paths=[file_path], metadata={ "preference": "personal" })
    mock_vector_db.sources = [file_source]
    mock_vector_db.query.return_value = [{"context": str_content_long, "score": 0.9}]

    query = "Where did Kuriko travel recently?"
    results = mock_vector_db.query(query)

    assert any("latin america" in result["context"].lower() for result in results)
    mock_vector_db.query.assert_called_once()


def test_multiple_short_file_knowledge_sources(mock_vector_db, tmpdir):
    contents = [
        {
            "content": "John works as a software engineer.",
            "metadata": {"category": "profession", "source": "occupation"},
        },
        {
            "content": "John lives in San Francisco.",
            "metadata": {"category": "city", "source": "personal"},
        },
        {
            "content": "John enjoys cooking Italian food.",
            "metadata": {"category": "hobby", "source": "personal"},
        },
    ]
    file_paths = []
    for i, item in enumerate(contents):
        file_path = Path(tmpdir.join(f"file_{i}.txt"))
        with open(file_path, "w") as f:
            f.write(item["content"])
        file_paths.append((file_path, item["metadata"]))

    file_sources = [TextFileKnowledgeSource(file_paths=[path], metadata=metadata) for path, metadata in file_paths]
    mock_vector_db.sources = file_sources
    mock_vector_db.query.return_value = [{"context": "John lives in San Francisco.", "score": 0.9}]

    query = "What city does he reside in?"
    results = mock_vector_db.query(query)

    assert any("san francisco" in result["context"].lower() for result in results)
    mock_vector_db.query.assert_called_once()


def test_multiple_2k_character_files(mock_vector_db, tmpdir):
    contents = [str_content_long * 2, str_content_long_2 * 2]
    file_paths = []
    for i, content in enumerate(contents):
        file_path = Path(tmpdir.join(f"long_file_{i}.txt"))
        with open(file_path, "w") as f:
            f.write(content)
        file_paths.append(file_path)

    file_sources = [TextFileKnowledgeSource(file_paths=[path], metadata={"preference": "personal"}) for path in file_paths]
    mock_vector_db.sources = file_sources
    mock_vector_db.query.return_value = [{ "context": contents[1], "score": 0.9 }]

    query = "What is J's cat name?"
    results = mock_vector_db.query(query)

    assert any("luna" in result["context"].lower() for result in results)
    mock_vector_db.query.assert_called_once()


# @pytest.mark.vcr(filter_headers=["authorization"])
def test_hybrid_string_and_file_knowledge_source(mock_vector_db, tmpdir):
    str_contents = [
        "Kuriko decided to take a stray kitten home.",
        "She enjoys quiet and peaceful morning.",
        "She named the kitten Luna."
    ]
    string_sources = [StringKnowledgeSource(content=item, metadata={ "preference": "personal" }) for item in str_contents]

    file_contents = ["J works on a computer all day.","J reads engineering journals."]
    file_paths = []
    for i, content in enumerate(file_contents):
        file_path = Path(tmpdir.join(f"file_{i}.txt"))
        with open(file_path, "w") as f:
            f.write(content)
        file_paths.append(file_path)
    file_sources = [TextFileKnowledgeSource(file_paths=[item], metadata={"preference": "personal"}) for item in file_paths]

    mock_vector_db.sources = string_sources + file_sources
    mock_vector_db.query.return_value = [{"context": file_contents[1], "score": 0.9 }]

    query = "What does J read?"
    results = mock_vector_db.query(query)

    assert any("engineering journals" in result["context"].lower() for result in results)
    mock_vector_db.query.assert_called_once()


def test_pdf_knowledge_source(mock_vector_db):
    current_dir = Path(__file__).parent
    pdf_path = current_dir / "mock_report_compressed.pdf"
    pdf_source = PDFKnowledgeSource(file_paths=[pdf_path], metadata={ "preference": "personal" })
    assert pdf_source.valid_file_paths == [pdf_path]
    assert pdf_source.content is not None

    mock_vector_db.sources = [pdf_source]
    mock_vector_db.query.return_value = [{"context": "McKinsey faces a challenge of manually curating and tagging documents, and takes hierarchical classification approach to query their internal knowledge base.", "score": 0.9} ]

    query = "What methodology does McKinsey take?"
    results = mock_vector_db.query(query)

    assert any("hierarchical classification" in result["context"].lower() for result in results)
    mock_vector_db.query.assert_called_once()


# @pytest.mark.vcr(filter_headers=["authorization"])
def test_csv_knowledge_source(mock_vector_db, tmpdir):
    csv_content = [
        ["Name", "Age", "City"],
        ["Kuriko", "55", "New York"],
        ["Alice", "30", "Los Angeles"],
        ["J", "60", "Chicago"],
    ]
    csv_path = Path(tmpdir.join("data.csv"))
    with open(csv_path, "w", encoding="utf-8") as f:
        for row in csv_content:
            f.write(",".join(row) + "\n")
    csv_source = CSVKnowledgeSource(file_paths=[csv_path], metadata={"preference": "personal"})
    mock_vector_db.sources = [csv_source]
    mock_vector_db.query.return_value = [{"context": "Kuriko is 55 years old.", "score": 0.9}]

    query = "How old is Kuriko?"
    results = mock_vector_db.query(query)

    assert any("55" in result["context"] for result in results)
    mock_vector_db.query.assert_called_once()


def test_excel_knowledge_source(mock_vector_db, tmpdir):
    import pandas as pd
    excel_data = {
        "Name": ["Kuriko", "Alice", "J"],
        "Age": [55, 30, 60],
        "City": ["New York", "Los Angeles", "Chicago"],
    }
    df = pd.DataFrame(excel_data)
    excel_path = Path(tmpdir.join("data.xlsx"))
    df.to_excel(excel_path, index=False)

    excel_source = ExcelKnowledgeSource(file_paths=[excel_path], metadata={"preference": "personal"})
    mock_vector_db.sources = [excel_source]
    mock_vector_db.query.return_value = [{ "context": "Kuriko is 55 years old.", "score": 0.9 }]

    query = "How old is Kuriko?"
    results = mock_vector_db.query(query)

    assert any("55" in result["context"] for result in results)
    mock_vector_db.query.assert_called_once()


def test_json_knowledge_source(mock_vector_db, tmpdir):
    json_data = {
        "people": [
            {"name": "Kuriko", "age": 55, "city": "New York"},
            {"name": "Alice", "age": 30, "city": "Los Angeles"},
            {"name": "J", "age": 60, "city": "Chicago"},
        ]
    }
    json_path = Path(tmpdir.join("data.json"))
    with open(json_path, "w", encoding="utf-8") as f:
        import json
        json.dump(json_data, f)

    json_source = JSONKnowledgeSource(file_paths=[json_path], metadata={ "preference": "personal" })
    mock_vector_db.sources = [json_source]
    mock_vector_db.query.return_value = [{"context": "Alice lives in Los Angeles.", "score": 0.9}]

    query = "Where does Alice reside?"
    results = mock_vector_db.query(query)

    assert any("los angeles" in result["context"].lower() for result in results)
    mock_vector_db.query.assert_called_once()


def test_docling_source(mock_vector_db):
    docling_source = DoclingSource(file_paths=["https://lilianweng.github.io/posts/2024-11-28-reward-hacking/",])
    mock_vector_db.sources = [docling_source]
    mock_vector_db.query.return_value = [
        {
            "context": "Reward hacking is a technique used to improve the performance of reinforcement learning agents.",
            "score": 0.9,
        }
    ]

    query = "What is reward hacking?"
    results = mock_vector_db.query(query)

    assert any("reward hacking" in result["context"].lower() for result in results)
    mock_vector_db.query.assert_called_once()


def test_multiple_docling_sources():
    urls: List[Path | str] = [
        "https://lilianweng.github.io/posts/2024-11-28-reward-hacking/",
        "https://lilianweng.github.io/posts/2024-07-07-hallucination/",
    ]
    docling_source = DoclingSource(file_paths=urls)

    assert docling_source.file_paths == urls
    assert docling_source.content is not None



from typing import List, Any
from typing_extensions import Self

from pydantic import BaseModel, model_validator
import pandas as pd
from sklearn.preprocessing import MinMaxScaler

from versionhq.memory.model import MemoryMetadata


class EvaluationItem(BaseModel):
    """
    A Pydantic class to store the evaluation result with scoring and suggestion based on the given criteria.
    This class will be used as a response format for the eval task.
    """
    criteria: str
    suggestion: str
    score: float
    weight: int


class Evaluation(BaseModel):
    """
    A Pydantic class to handle evaluation of the task output.
    """

    items: List[EvaluationItem] = []
    eval_by: Any = None


    @model_validator(mode="after")
    def set_up_evaluator(self) -> Self:
        from versionhq.agent.inhouse_agents import vhq_task_evaluator
        self.eval_by = vhq_task_evaluator
        return self


    def _create_memory_metadata(self) -> MemoryMetadata:
        """
        Create and store evaluation results in the memory metadata
        """
        eval_by = self.eval_by.key # saving memory
        score = self.aggregate_score
        eval_criteria = ", ".join([item.criteria for item in self.items]) if self.items else None
        suggestion = self.suggestion_summary
        memory_metadata = MemoryMetadata(eval_by=eval_by, score=score, eval_criteria=eval_criteria, suggestion=suggestion)
        return memory_metadata


    def _draft_fsl_prompt(self, task_description: str = None) -> str | None:
        """
        Search competitive and weak cases in the past and draft few shot learning prompt.
        """
        from versionhq.task.TEMPLATES.Description import SHOTS
        shot_prompt = None

        if self.eval_by.long_term_memory:
            res = self.eval_by.long_term_memory.search(query=task_description, latest_n=10)

            if res:
                new_res = filter(lambda x: "score" in x["metadata"], res)
                new_res = list(new_res)
                new_res.sort(key=lambda x: x["metadata"]["score"], reverse=True)
                if new_res and new_res[0]['data']:
                    c = new_res[0]['data']['task_output']
                    w = new_res[len(new_res)-1]['data']['task_output'] if new_res[len(new_res)-1]['metadata']['score'] < new_res[0]['metadata']['score'] else ""
                    shot_prompt = SHOTS.format(c=c, w=w)

        return shot_prompt


    def _normalize_df(self) -> pd.DataFrame:
        """
        Creates a pandas DataFrame from a list of EvaluationItem objects containing 'weight' and 'score' columns, and normalizes them using MinMaxScaler.

        Args:
            items: A list of EvaluationItem objects.

        Returns:
            A pandas DataFrame with normalized 'weight' and 'score' columns, or an empty DataFrame if the input is empty.
        """

        if not self.items:
            return pd.DataFrame()

        data = { 'weight': [item.weight for item in self.items], 'score': [item.score for item in self.items] }
        df = pd.DataFrame(data)

        scaler = MinMaxScaler(feature_range=(0, 1))
        df[['weight', 'score']] = scaler.fit_transform(df[['weight', 'score']])
        return df


    @property
    def aggregate_score(self) -> int | float:
        if not self.items:
            return 0

        df = self._normalize_df()
        df['weighted_score'] = df['weight'] * df['score']
        n = df['weighted_score'].sum()
        if n == 0.0 or n == 1.0:
            import math
            s = [[item.score for item in self.items]]
            w = [[item.weight for item in self.items]]
            r = [math.sumprod(x, y) for x, y in zip(s, w)]
            if r and sum(w[0]):
                n = r[0] / sum(w[0])

        aggregate_score = round(n, 3)
        return aggregate_score


    @property
    def suggestion_summary(self) -> str | None:
        """Returns a summary of the suggestions"""

        if not self.items:
            return None

        summary = ""
        for item in self.items:
            summary += f"{item.suggestion}, "

        return summary



from versionhq._utils.logger import Logger
from versionhq._utils.process_config import process_config
from versionhq._utils.vars import KNOWLEDGE_DIRECTORY, MAX_FILE_NAME_LENGTH
from versionhq._utils.is_valid_url import is_valid_url
from versionhq._utils.usage_metrics import UsageMetrics, ErrorType
from versionhq._utils.convert_img_url import convert_img_url
from versionhq._utils.is_valid_enum import is_valid_enum
from versionhq._utils.handle_directory import handle_directory



import re
from playwright.sync_api import Page, expect

def _test_has_title(page: Page):
    page.goto("https://playwright.dev/")
    expect(page).to_have_title(re.compile("Playwright"))


def _test_get_started_link(page: Page):
    page.goto("https://playwright.dev/")
    page.get_by_role("link", name="Get started").click()

    expect(page.get_by_role("heading", name="Installation")).to_be_visible()



EVALUATE="""Evaluate the provided task output against the given task description, assigning a score between 0 (worst) and 1 (best) based on the specified criteria. Scores should be numerical (integers or decimals). Weight should be numerical (integers or decimals) and represents importance of the criteria to the final result. Provide specific suggestions for improvement. Do not assign identical scores to different criteria unless otherwise you have clear reasons to do so:
Task output: {task_output}
Task description: {task_description}
Evaluation criteria: {eval_criteria}
"""

SHOTS="""Here are two examples of task outputs. The first is considered excellent due to its clear planning and alignment with the goal. The second is weak due to clichéd language. Now, evaluate the given task output.
First = Excellent example: {c}
Second = Weak example: {w}
"""



import re
import gzip
import http.client
import urllib.request
from urllib.request import Request
from textwrap import dedent
from typing import Any, Optional, List, Dict

from pydantic import Field

from versionhq.agent.model import Agent
from versionhq.tool.model import BaseTool
from versionhq._utils.logger import Logger



class RagTool(BaseTool):
    """A Pydantic class to store a RAG tool object. Inherited from BaseTool"""

    api_key_name: str = Field(default=None)
    api_endpoint: Optional[str] = Field(default=None)

    url: Optional[str] = Field(default=None, description="url to scrape")
    headers: Optional[Dict[str, Any]] = Field(default_factory=dict, description="request headers")

    sources: Optional[List[Any]] = Field(default_factory=list, description="indexed data sources")
    query: Optional[str] = Field(default=None)
    text: Optional[str] = Field(default=None, description="text data source")


    def _sanitize_source_code(self, source_code: str | bytes = None) -> str | None:
        if not source_code:
            return None

        if isinstance(source_code, bytes):
            source_code = source_code.decode('utf-8')

        try:
            import html2text
        except:
            Logger().log(message="Dependencies for tools are missing. Add tool packages by running: `uv add versionhq[tool]`.", color="red", level="error")

        h = html2text.HTML2Text()
        h.ignore_links = False
        text = h.handle(source_code)
        text = re.sub(r"[^a-zA-Z$0-9\s\n]", "", text)
        return dedent(text)


    def _scrape_url(self, url: str = None) -> str | None:
        url = url if url else self.url

        if not url:
            return None

        http.client.HTTPConnection.debuglevel = 1

        try:
            req = Request(url=url, headers=self.headers, origin_req_host=url, method="GET")
            res = ""

            with urllib.request.urlopen(req) as url:
                if url.info().get("Content-Encoding") == "gzip":
                    res = gzip.decompress(url.read())
                else:
                    res = url.read()

            text = self._sanitize_source_code(source_code=res)
            return text

        except urllib.request.HTTPError as e:
            Logger().log(level="error", message=f"HTTP error occurred: {str(e)}", color="red")
            return None

        except Exception as e:
            Logger().log(level="error", message=f"Error fetching URL {self.api_endpoint}: {str(e)}", color="red")
            return None


    def store_data(self, agent: Agent = None) -> None:
        """Stores retrieved data in the storage"""
        if not agent:
            return

        text = self.text if self.text else self._scrape_url(self.url)
        self.text = text
        knowledge_sources = [*agent.knowledge_sources, str(text), ] if agent.knowledge_sources else [str(text),]
        agent.update(knowledge_sources=knowledge_sources)


    def _run(self, agent: Agent = None, query: str = None) -> List[str]:
        query = query if query else self.query

        if not query or not agent:
            text = self.text if self.text else self._scrape_url(self.url)
            self.text = text
            return [text,]

        else:
            results, res = [], []
            if agent._knowledge:
                res = agent._knowledge.query(query=[query], limit=5)
            else:
                self.store_data(agent=agent)
                res = agent._knowledge.query(query=[query], limit=5)

            for item in res:
                if isinstance(item, dict):
                    results.append(item["context"])
                else:
                    results.append(str(item))
            return results


    def run(self, *args, **kwargs):
        return self._run(*args, **kwargs)



import csv
import json
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional
from pathlib import Path

import numpy as np
from pydantic import BaseModel, ConfigDict, Field, field_validator

from versionhq.knowledge.storage import KnowledgeStorage
from versionhq.storage.utils import fetch_db_storage_path
from versionhq._utils.vars import KNOWLEDGE_DIRECTORY
from versionhq._utils.logger import Logger


class BaseKnowledgeSource(BaseModel, ABC):
    """
    Abstract base class for knowledge sources: csv, json, excel, pdf, string, and docling.
    """
    _logger: Logger = Logger(verbose=True)

    chunk_size: int = 3000
    chunk_overlap: int = 200
    chunks: List[str] = Field(default_factory=list)
    chunk_embeddings: List[np.ndarray] = Field(default_factory=list)

    model_config = ConfigDict(arbitrary_types_allowed=True)
    storage: Optional[KnowledgeStorage] = Field(default=None)
    metadata: Dict[str, Any] = Field(default_factory=dict)
    collection_name: Optional[str] = Field(default=None)


    @abstractmethod
    def validate_content(self, **kwargs) -> Any:
        """Load and preprocess content from the source."""
        pass


    @abstractmethod
    def add(self) -> None:
        """Process content, chunk it, compute embeddings, and save them."""
        pass


    def get_embeddings(self) -> List[np.ndarray]:
        """Return the list of embeddings for the chunks."""
        return self.chunk_embeddings


    def _chunk_text(self, text: str) -> List[str]:
        """
        Utility method to split text into chunks.
        """
        return [text[i : i + self.chunk_size] for i in range(0, len(text), self.chunk_size - self.chunk_overlap)]


    def _save_documents(self) -> None:
        """
        Save the documents to the given (or newly created) storage on ChromaDB.
        This method should be called after the chunks and embeddings are generated.
        """
        # if not self.chunks or self.chunk_embeddings:
        #     self._logger.log(level="warning", message="Chunks or chunk embeddings are missing. Save docs after creating them.", color="yellow")
        #     return

        try:
            if self.storage:
                self.storage.save(documents=self.chunks, metadata=self.metadata)

            else:
                storage = KnowledgeStorage(collection_name=self.collection_name) if self.collection_name else KnowledgeStorage()
                storage.initialize_knowledge_storage()
                self.storage = storage
                self.storage.save(documents=self.chunks, metadata=self.metadata)

        except:
            self._logger.log(level="error", message="No storage found or created to save the documents.", color="red")
            return
            # raise ValueError("No storage found to save documents.")



class StringKnowledgeSource(BaseKnowledgeSource):
    """
    A knowledge source that stores and queries plain text content using embeddings.
    """

    content: str = Field(...)
    collection_name: Optional[str] = Field(default=None)

    def model_post_init(self, _):
        """Post-initialization method to validate content."""
        self.validate_content()
        self._save_documents()


    def validate_content(self):
        """Validate string content."""
        if not isinstance(self.content, str):
            raise ValueError("StringKnowledgeSource only accepts string content")


    def add(self) -> None:
        """
        Add string content to the knowledge source, chunk it, compute embeddings, and save them.
        """
        new_chunks = self._chunk_text(text=self.content)
        self.chunks.extend(new_chunks)
        self._save_documents()



class BaseFileKnowledgeSource(BaseKnowledgeSource, ABC):
    """Base class for knowledge sources that load content from files."""

    file_paths: Optional[Path | List[Path] | str | List[str]] = Field(default_factory=list)
    content: Dict[Path, str] = Field(init=False, default_factory=dict)
    storage: Optional[KnowledgeStorage] = Field(default=None)
    valid_file_paths: List[Path] = Field(default_factory=list, description="store a list of `Path` objects from self.file_paths")


    @field_validator("file_paths", mode="before")
    def validate_file_path(cls, v, info):
        """
        Validate if at least one valid file path is provided.
        """
        if v is None and info.data.get("file_paths") is None:
            raise ValueError("Either file_path or file_paths must be provided")
        return v


    def validate_content(self, path: str | Path) -> List[Path]:
        """
        Convert the given path to a Path object, and validate if the path exists and refers to a file.)
        """

        path_instance = Path(KNOWLEDGE_DIRECTORY + "/" + path) if isinstance(path, str) else path

        if not path_instance.exists():
            abs_path = fetch_db_storage_path()
            path_instance = Path(abs_path + "/" + KNOWLEDGE_DIRECTORY + "/" + path) if isinstance(path, str) else path

            if not path_instance.exists():
                self._logger.log(level="error", message="File path not found.", color="red")
                raise ValueError()

            elif not path_instance.is_file():
                self._logger.log(level="error", message="Non-file object was given.", color="red")
                raise ValueError()

        elif not path_instance.is_file():
            self._logger.log(level="error", message="Non-file object was given.", color="red")
            raise ValueError()

        return path_instance



    def _process_file_paths(self) -> List[Path]:
        """
        Convert file_path to a list of Path objects.
        """
        if not self.file_paths:
            self._logger.log(level="error", message="Missing file paths.", color="red")
            raise ValueError("Missing file paths.")


        path_list: List[Path | str] = [self.file_paths] if isinstance(self.file_paths, (str, Path)) else list(self.file_paths) if isinstance(self.file_paths, list) else []
        valid_path_list = list()

        if not path_list:
            self._logger.log(level="error", message="Missing valid file paths.", color="red")
            raise ValueError("Your source must be provided with file_paths: []")

        for item in path_list:
            valid_path = self.validate_content(item)
            if valid_path:
                valid_path_list.append(valid_path)

        return valid_path_list


    def model_post_init(self, _) -> None:
        """
        Post-initialization method to load content.
        """
        self.valid_file_paths = self._process_file_paths()
        self.content = self.load_content()
        self._save_documents()


    @abstractmethod
    def load_content(self) -> Dict[Path, str]:
        """
        Load and preprocess file content. Should be overridden by subclasses.
        Assume that the file path is relative to the project root in the knowledge directory.
        """
        pass



class TextFileKnowledgeSource(BaseFileKnowledgeSource):
    """
    A knowledge source class that stores and queries text file content using embeddings.
    """

    def load_content(self) -> Dict[Path, str]:
        """
        Load and preprocess text file content.
        """
        content = {}
        for path in self.valid_file_paths:
            path = self.validate_content(path=path)
            with open(path, "r", encoding="utf-8") as f:
                content[path] = f.read()
        return content


    def add(self) -> None:
        """
        Add text file content to the knowledge source, chunk it, compute embeddings, and save the embeddings.
        """
        for _, text in self.content.items():
            new_chunks = self._chunk_text(text=text)
            self.chunks.extend(new_chunks)

        self._save_documents()



class PDFKnowledgeSource(BaseFileKnowledgeSource):
    """
    A knowledge source class that stores and queries PDF file content using embeddings.
    """

    def load_content(self) -> Dict[Path, str]:
        """
        Load and preprocess PDF file content.
        """
        self._import_pdfplumber()

        import pdfplumber

        content = {}
        for path in self.valid_file_paths:
            text = ""
            path = self.validate_content(path)
            with pdfplumber.open(path) as pdf:
                for page in pdf.pages:
                    page_text = page.extract_text()
                    if page_text:
                        text += page_text + "\n"
            content[path] = text
        return content


    def _import_pdfplumber(self):
        """
        Dynamically import pdfplumber.
        """
        try:
            import pdfplumber
        except ImportError:
            try:
                import os
                os.system("uv add pdfplumber --optional pdfplumber")
            except:
                raise ImportError("pdfplumber is not installed. Please install it with: $ uv add versionhq[pdfplumber]")


    def add(self) -> None:
        """
        Add PDF file content to the knowledge source, chunk it, compute embeddings, and save the embeddings.
        """
        for _, text in self.content.items():
            new_chunks = self._chunk_text(text=text)
            self.chunks.extend(new_chunks)

        self._save_documents()



class CSVKnowledgeSource(BaseFileKnowledgeSource):
    """
    A knowledge source class that stores and queries CSV file content using embeddings.
    """

    def load_content(self) -> Dict[Path, str]:
        """
        Load and preprocess CSV file content.
        """
        content_dict = {}
        for file_path in self.valid_file_paths:
            with open(file_path, "r", encoding="utf-8") as csvfile:
                reader = csv.reader(csvfile)
                content = ""
                for row in reader:
                    content += " ".join(row) + "\n"
                content_dict[file_path] = content

        return content_dict


    def add(self) -> None:
        """
        Add CSV file content to the knowledge source, chunk it, compute embeddings, and save the embeddings.
        """
        content_str = str(self.content) if isinstance(self.content, dict) else self.content
        new_chunks = self._chunk_text(text=content_str)
        self.chunks.extend(new_chunks)
        self._save_documents()



class JSONKnowledgeSource(BaseFileKnowledgeSource):
    """
    A knowledge source class that stores and queries JSON file content using embeddings.
    """

    def load_content(self) -> Dict[Path, str]:
        """
        Load and preprocess JSON file content.
        """
        content: Dict[Path, str] = {}
        for path in self.valid_file_paths:
            path = self.validate_content(path)
            with open(path, "r", encoding="utf-8") as json_file:
                data = json.load(json_file)
            content[path] = self._json_to_text(data)
        return content


    def _json_to_text(self, data: Any, level: int = 0) -> str:
        """
        Recursively convert JSON data to a text representation.
        """
        text = ""
        indent = "  " * level
        if isinstance(data, dict):
            for key, value in data.items():
                text += f"{indent}{key}: {self._json_to_text(value, level + 1)}\n"
        elif isinstance(data, list):
            for item in data:
                text += f"{indent}- {self._json_to_text(item, level + 1)}\n"
        else:
            text += f"{str(data)}"
        return text


    def add(self) -> None:
        """
        Add JSON file content to the knowledge source, chunk it, compute embeddings, and save the embeddings.
        """
        content_str = str(self.content) if isinstance(self.content, dict) else self.content
        new_chunks = self._chunk_text(text=content_str)
        self.chunks.extend(new_chunks)
        self._save_documents()



class ExcelKnowledgeSource(BaseFileKnowledgeSource):
    """
    A knowledge source that stores and queries Excel file content using embeddings.
    """

    def load_content(self) -> Dict[Path, str]:
        """
        Load and preprocess Excel file content.
        """

        pd = self._import_dependencies()
        content_dict = {}
        for file_path in self.valid_file_paths:
            file_path = self.validate_content(file_path)
            df = pd.read_excel(file_path)
            content = df.to_csv(index=False)
            content_dict[file_path] = content
        return content_dict


    def _import_dependencies(self):
        """
        Dynamically import dependencies.
        """
        try:
            import pandas as pd
            return pd
        except ImportError as e:
            try:
                import os
                os.system("uv add pandas --optional pandas")
                import pandas as pd
                return pd
            except:
                missing_package = str(e).split()[-1]
                raise ImportError(
                    f"{missing_package} is not installed. Please install it with: $ uv add versionhq[{missing_package}]"
                )


    def add(self) -> None:
        """
        Add Excel file content to the knowledge source, chunk it, compute embeddings, and save the embeddings.
        """
        content_str = "\n".join(str(value) for value in self.content.values()) if isinstance(self.content, dict) else str(self.content)
        new_chunks = self._chunk_text(text=content_str)
        self.chunks.extend(new_chunks)
        self._save_documents()



from urllib.parse import urlparse


def is_valid_url(url: str) -> bool:
    try:
        result = urlparse(url)
        return all(
            [
                result.scheme in ("http", "https", "gs"),
                result.netloc,
                len(result.netloc.split(".")) >= 2,  # Ensure domain has TLD
            ]
        )
    except Exception:
        return False



import threading
import time
from typing import Optional
from typing_extensions import Self

from pydantic import BaseModel, Field, PrivateAttr, model_validator

from versionhq._utils import Logger


class RPMController(BaseModel):
    max_rpm: Optional[int] = Field(default=None)
    _logger: Logger = PrivateAttr(default_factory=lambda: Logger(verbose=True))
    _current_rpm: int = PrivateAttr(default=0)
    _timer: Optional[threading.Timer] = PrivateAttr(default=None)
    _lock: Optional[threading.Lock] = PrivateAttr(default=None)
    _shutdown_flag: bool = PrivateAttr(default=False)

    @model_validator(mode="after")
    def reset_counter(self) -> Self:
        if self.max_rpm is not None:
            if not self._shutdown_flag:
                self._lock = threading.Lock()
                self._reset_request_count()
        return self


    def _check_and_increment(self) -> bool:
        if self.max_rpm is None:
            return True

        elif self.max_rpm is not None and self._current_rpm < self.max_rpm:
            self._current_rpm += 1
            return True

        elif self.max_rpm is not None and self._current_rpm >= self.max_rpm:
            self._logger.log(level="info", message="Max RPM reached, waiting for next minute to start.", color="yellow")
            self._wait_for_next_minute()
            self._current_rpm = 1 # restart
            return True

        else:
            return False

    def check_or_wait(self) -> bool:
        if self._lock:
            with self._lock:
                return self._check_and_increment()
        else:
            return self._check_and_increment()

        return False


    def stop_rpm_counter(self) -> None:
        if self._timer:
            self._timer.cancel()
            self._timer = None


    def _wait_for_next_minute(self) -> None:
        time.sleep(60)
        self._current_rpm = 0


    def _reset_request_count(self) -> None:
        def _reset():
            self._current_rpm = 0
            if not self._shutdown_flag:
                self._timer = threading.Timer(60.0, self._reset_request_count)
                self._timer.start()

        if self._lock:
            with self._lock:
                _reset()
        else:
            _reset()

        if self._timer:
            self._shutdown_flag = True
            self._timer.cancel()



"""Test code snippets on docs/core/task.md"""

from versionhq._utils import UsageMetrics
from versionhq._prompt.auto_feedback import PromptFeedbackGraph


def test_docs_core_task_a():
    import versionhq as vhq

    task = vhq.Task(description="MY AMAZING TASK")

    import uuid
    assert uuid.UUID(str(task.id), version=4)


def test_docs_core_task_b():
    import versionhq as vhq

    task = vhq.Task(description="MY AMAZING TASK")
    res = task.execute()

    assert isinstance(res, vhq.TaskOutput)
    assert res.raw and res.json
    assert task.processed_agents is not None


def test_docs_core_task_c():
    import versionhq as vhq
    from tests import Demo

    task = vhq.Task(
        description="generate random output that strictly follows the given format",
        response_schema=Demo,
    )
    res = task.execute()

    assert isinstance(res, vhq.TaskOutput) and res.task_id is task.id
    assert isinstance(res.raw, str) and isinstance(res.json_dict, dict)
    assert [
        getattr(res.pydantic, k) is not None and v.annotation == Demo.model_fields[k].annotation
        for k, v in res.pydantic.model_fields.items()
    ]


def test_docs_core_task_d():
    import versionhq as vhq
    from tests import demo_response_fields

    task = vhq.Task(
        description="Output random values strictly following the data type defined in the given response format.",
        response_schema=demo_response_fields
    )
    res = task.execute()

    assert isinstance(res, vhq.TaskOutput) and res.task_id is task.id
    assert res.raw is not None and res.json_dict is not None
    assert res.pydantic is None
    assert [v and type(v) == task.response_schema[i].data_type for i, (k, v) in enumerate(res.json_dict.items())]


def test_docs_core_task_e():
    from typing import Any

    from pydantic import BaseModel
    import versionhq as vhq

    # 1. Define and execute a sub task with Pydantic output.
    class Sub(BaseModel):
        sub1: str
        sub2: dict[str, Any]

    sub_task = vhq.Task(description="generates a random value that strictly follows the given format.", response_schema=Sub)
    sub_res = sub_task.execute()

    class Main(BaseModel):
        main1: list[Any]
        main2: str

    def format_response(sub, **kwargs) -> Main:
        main1 = kwargs["main1"] if kwargs and "main1" in kwargs else None
        if main1:
            main1.append(sub)
        main2 = kwargs["main2"] if kwargs and "main2" in kwargs else None
        main = Main(main1=main1, main2=str(main2))
        return main

    main_task = vhq.Task(
        description="generates a random value that strictly follows the givne format.",
        response_schema=Main,
        callback=format_response,
        callback_kwargs=dict(sub=sub_res.json_dict),
    )
    res = main_task.execute(context=sub_res.raw) # [Optional] Adding sub_task's response as context.
    assert res.callback_output.main1 is not None
    assert res.callback_output.main2 is not None


def test_docs_core_task_g():
    import versionhq as vhq
    task = vhq.Task(
        description="return the output following the given prompt.",
        allow_delegation=True
    )
    task.execute()

    assert task.output is not None
    assert task.processed_agents is not None
    assert task._delegations ==1


def test_docs_core_task_h():
    import versionhq as vhq

    task = vhq.Task(description="Return a word: 'test'", type=vhq.TaskExecutionType.ASYNC)

    from unittest.mock import patch
    with patch.object(vhq.Agent, "execute_task", return_value=("user prompt", "dev prompt", "test", UsageMetrics())) as execute:
        res = task.execute()
        assert res.raw == "test"
        execute.assert_called_once_with(task=task, context=None)


def test_docs_core_task_i():
    import versionhq as vhq

    def random_func(message: str) -> str:
        return message + "_demo"

    tool = vhq.Tool(name="tool", func=random_func)
    tool_set = vhq.ToolSet(tool=tool, kwargs=dict(message="empty func"))
    task = vhq.Task(
        description="execute the given tools",
        tools=[tool_set,],
        tool_res_as_final=True
    )

    res = task.execute()
    assert res.tool_output == "empty func_demo"


def test_docs_core_task_j():
    import versionhq as vhq

    simple_tool = vhq.Tool(name="simple tool", func=lambda x: "simple func")
    agent = vhq.Agent(role="demo", goal="execute tools", tools=[simple_tool,])
    task = vhq.Task(
        description="execute tools",
        can_use_agent_tools=True,
        tool_res_as_final=True
    )
    res = task.execute(agent=agent)
    assert res.tool_output == "simple func"


def test_docs_core_task_k():
    import versionhq as vhq

    def callback_func(condition: str, test1: str, **kwargs):
        return f"Result: {test1}, condition added: {condition}"

    task = vhq.Task(
        description="return the output following the given prompt.",
        callback=callback_func,
        callback_kwargs=dict(condition="demo for pytest")
    )
    res = task.execute()

    assert res and isinstance(res, vhq.TaskOutput)
    assert res.task_id is task.id
    assert "demo for pytest" in res.callback_output


def test_docs_core_task_l():
    import versionhq as vhq
    from pathlib import Path

    current_dir = Path(__file__).parent.parent
    file_path = current_dir / "_sample/screenshot.png"
    audio_path = current_dir / "_sample/sample.mp3"

    task = vhq.Task(description="Summarize the given content", image=str(file_path), audio=str(audio_path))
    res = task.execute(agent=vhq.Agent(llm="gemini-2.0", role="Content Interpretator"))

    assert res.final is not None


def test_docs_core_task_m():
    import versionhq as vhq

    task = vhq.Task(description="Create a short story.", should_test_run=True, human=False)
    task.execute()
    assert isinstance(task._pfg, PromptFeedbackGraph)
    assert task._pfg.should_reform == False
    assert task._pfg.reform_trigger_event == None
    assert [k for k in task._pfg.user_prompts.keys()] and [k for k in task._pfg.dev_prompts.keys()]
    assert task._pfg.usage.total_tokens == task.output.usage.total_tokens
    assert task.output.usage.total_errors == task._pfg.usage.total_errors
    assert task.output.usage.latency == task._pfg.usage.latency



from enum import Enum, IntEnum
from typing import Any


def is_valid_enum(enum: Enum | IntEnum, key: str = None, val: str | Enum | IntEnum = None) -> bool:
    if not enum: return False

    if key:
        key = key.upper()
        matched = [k for k in enum._member_map_.keys() if hasattr(enum, "_member_map_") and k == key]
        return bool(matched)

    elif val:
        match val:
            case str():
                matched = [k for k in enum._value2member_map_.keys() if hasattr(enum, "_value2member_map_") and k == val]
                return bool(matched)

            case Enum() | IntEnum():
                return val in enum

            case _:
                return False

    else: return False



import datetime
from typing import Dict, Any, Optional, Tuple, List

from versionhq.tool.gpt import openai_client
from versionhq.tool.gpt._enum import GPTSizeEnum
from versionhq._utils import is_valid_enum, UsageMetrics, ErrorType


class GPTToolWebSearch:
    """A class to manage Web Search tools by OpenAI."""

    model: str = "gpt-4o"
    input: str = None
    location_type: str = None  # "approximate"
    country: str = None  # "GB"
    city: str = None # "London"
    region: str = None # "London"
    search_content_size: str = GPTSizeEnum.MEDIUM.value
    _user_location: Optional[Dict[str, str]] = None
    _usage: UsageMetrics = UsageMetrics()


    def __init__(
            self,
            model: str = None,
            input: str = None,
            location_type: str = None,
            country: str = None,
            city: str = None,
            region: str = None,
            search_content_size = str | GPTSizeEnum,
            **kwargs,
        ):
        self.model = model if model else self.model
        self.input = input if input else self.input
        if country and city and region:
            self.location_type = location_type if location_type else "approximate"
            self.country = country
            self.city = city
            self.region = region
            self._user_location = dict(type=self.location_type, country=self.country, city=self.city, region=self.region)

        if is_valid_enum(val=search_content_size, enum=GPTSizeEnum):
            self.search_content_size = search_content_size if isinstance(search_content_size, str) else search_content_size.value

        if kwargs:
            for k, v in kwargs.items():
                if hasattr(self, k):
                    setattr(self, k, v)


    def run(self) -> Tuple[str, List[Dict[str, str]], UsageMetrics]:
        """Runs the tool and returns text response, annotations, and usage metrics."""

        raw_res = ""
        annotations = list()
        usage = self._usage if self._usage else UsageMetrics()
        start_dt = datetime.datetime.now()

        try:
            res = openai_client.responses.create(**self.schema)
            if not res:
                usage.record_errors(ErrorType.TOOL)
            else:
                content = res.output[0].content[0] if len(res.output) == 1 else res.output[1].content[0]
                if content:
                    raw_res = content.text
                    annotations = [{ "title": item.title, "url": item.url } for item in content.annotations] if content.annotations else []
                    usage.record_token_usage(**res.usage.__dict__)
        except:
            usage.record_errors(ErrorType.TOOL)
        end_dt = datetime.datetime.now()
        usage.record_latency(start_dt=start_dt, end_dt=end_dt)
        self._usage = usage
        return raw_res, annotations, usage


    @property
    def tool_schema(self) -> Dict[str, Any]:
        if self._user_location:
            return {
                "type": "web_search_preview",
                "user_location": self._user_location,
                "search_context_size": self.search_content_size
            }
        else:
            return {
                "type": "web_search_preview",
                "search_context_size": self.search_content_size
            }


    @property
    def schema(self) -> Dict[str, Any]:
        schema = dict(model=self.model, tools=[self.tool_schema,], input=self.input)
        return schema



from pydantic import BaseModel

from versionhq._prompt.model import Prompt
from versionhq.task_graph.model import ReformTriggerEvent
from versionhq._prompt.auto_feedback import PromptFeedbackGraph
from versionhq.agent.model import Agent
from versionhq.task.model import Task, TaskOutput


def test_pfg():
    class Travel(BaseModel):
        schedule: str
        destination: str
        other_consideration: str

    main_task = Task(description="plan a day trip to Costa Rica", response_schema=Travel)
    agent = Agent(llm="gemini-2.0", role="Demo Agent", maxit=1)
    prompt  = Prompt(task=main_task, agent=agent, context=["test", "test2"])
    pfg = PromptFeedbackGraph(prompt=prompt)
    assert pfg.prompts["0"] == prompt
    assert pfg._times_iteration == 0
    assert pfg.user_prompts["0"] is not None
    assert pfg.dev_prompts["0"] is not None

    pfg.set_up_graph(should_reform=False, reform_trigger_event=ReformTriggerEvent.ERROR_DETECTION)
    assert len([k for k in pfg.nodes.keys()]) == 5
    assert pfg.edges is not None
    assert pfg.should_reform == False
    assert pfg.reform_trigger_event == ReformTriggerEvent.ERROR_DETECTION

    res, all_outputs = pfg.activate()
    assert isinstance(res, TaskOutput)
    assert res.pydantic.schedule and res.pydantic.destination and res.pydantic.other_consideration
    assert all_outputs
